"""
Custom Django Admin views used in enterprise app.
"""

import datetime
import logging

from rest_framework.response import Response
from rest_framework.views import APIView

from django.conf import settings
from django.contrib import admin, auth, messages
from django.contrib.auth import get_permission_codename
from django.core.exceptions import ValidationError
from django.core.management import call_command
from django.db import transaction
from django.db.models import Q
from django.http import HttpResponse, HttpResponseForbidden, HttpResponseRedirect, HttpResponseServerError, JsonResponse
from django.shortcuts import get_object_or_404, render
from django.urls import reverse
from django.utils.translation import gettext as _
from django.utils.translation import ngettext
from django.views.generic import View

from enterprise.admin.forms import (
    ManageLearnersDataSharingConsentForm,
    ManageLearnersForm,
    TransmitEnterpriseCoursesForm,
)
from enterprise.admin.utils import (
    UrlNames,
    email_or_username__to__email,
    paginated_list,
    parse_csv,
    split_usernames_and_emails,
)
from enterprise.api_client.discovery import get_course_catalog_api_service_client
from enterprise.api_client.ecommerce import EcommerceApiClient
from enterprise.api_client.sso_orchestrator import EnterpriseSSOOrchestratorApiClient, SsoOrchestratorClientError
from enterprise.constants import DJANGO_ADMIN_MANAGE_LEARNERS_LIMIT, PAGE_SIZE
from enterprise.errors import LinkUserToEnterpriseError
from enterprise.models import (
    EnrollmentNotificationEmailTemplate,
    EnterpriseCatalogQuery,
    EnterpriseCustomer,
    EnterpriseCustomerUser,
    PendingEnterpriseCustomerUser,
)
from enterprise.utils import (
    ValidationMessages,
    delete_data_sharing_consent,
    enroll_users_in_course,
    get_ecommerce_worker_user,
    get_sso_orchestrator_configure_edx_oauth_path,
    validate_course_exists_for_enterprise,
    validate_email_to_link,
)

# Only create manual enrollments if running in edx-platform
try:
    from common.djangoapps.student.api import create_manual_enrollment_audit
except ImportError:
    create_manual_enrollment_audit = None
User = auth.get_user_model()

LOG = logging.getLogger(__name__)


class TemplatePreviewView(View):
    """
    Renders a given NotificationTemplate object to HTML for online viewing.
    """
    view_type_contexts = {
        "course": {
            "enrolled_in": {
                "name": "OpenEdX Demo Course",
                "url": "http://example.com/courses/edx-demo-course",
                "type": "course",
                "start": datetime.datetime.strptime('2016-01-01', '%Y-%m-%d')
            },
            "organization_name": "OpenEdX",
        },
        "program": {
            "enrolled_in": {
                "name": "OpenEdX Demo Program",
                "url": "http://example.com/programs/edx-demo-program",
                "type": "program",
                "branding": "MicroMasters",
                "start": datetime.datetime.strptime('2016-01-01', '%Y-%m-%d')
            },
            "organization_name": "OpenEdX",
        },
    }

    def get(self, request, template_id, view_type):
        """
        Render the given template with the stock data.
        """
        template = get_object_or_404(EnrollmentNotificationEmailTemplate, pk=template_id)
        if view_type not in self.view_type_contexts:
            return HttpResponse(status=404)
        base_context = self.view_type_contexts[view_type].copy()
        base_context.update({'user_name': self.get_user_name(request)})
        return HttpResponse(template.render_html_template(base_context), content_type='text/html')

    @staticmethod
    def get_user_name(request):
        """
        Get a human-readable name for the user.
        """
        return request.user.first_name or request.user.username


class CatalogQueryPreviewView(APIView):
    """
    Renders a search/all response from Discovery for a particular catalog query.
    """

    def get(self, request, catalog_query_id):
        """
        Render the response for a particular catalog query.
        """
        enterprise_catalog_query = get_object_or_404(EnterpriseCatalogQuery, pk=catalog_query_id)
        return Response(self.get_response_from_discovery(request, enterprise_catalog_query))

    @staticmethod
    def get_response_from_discovery(request, catalog_query):
        """
        POST a query to Discovery's search/all endpoint and return the results.
        """
        response = get_course_catalog_api_service_client().get_catalog_results(
            catalog_query.content_filter,
            traverse_pagination=True
        )
        return response


class BaseEnterpriseCustomerView(View):
    """
    Base class for Enterprise Customer views.
    """
    template = None

    @staticmethod
    def _get_admin_context(request, customer):
        """
        Build common admin context.
        """
        opts = customer._meta
        codename = get_permission_codename('change', opts)
        has_change_permission = request.user.has_perm('{}.{}'.format(opts.app_label, codename))
        return {
            'has_change_permission': has_change_permission,
            'opts': opts
        }

    def _get_view_context(self, request, customer_uuid):
        """
        Return the default context parameters
        """
        return {
            'enterprise_customer': EnterpriseCustomer.objects.get(uuid=customer_uuid)
        }

    def _build_context(self, request, customer_uuid):
        """
        Build common context parts used by different handlers in this view.
        """
        context = self._get_view_context(request, customer_uuid)
        context.update(admin.site.each_context(request))
        context.update(self._get_admin_context(request, context['enterprise_customer']))
        return context

    def get_form_view(self, request, customer_uuid, additional_context=None):
        """
        render the form with appropriate context.
        """
        context = self._build_context(request, customer_uuid)
        if additional_context:
            context.update(additional_context)
        return render(request, self.template, context)


class EnterpriseCustomerManageLearnerDataSharingConsentView(BaseEnterpriseCustomerView):
    """
    Manage Learners Data Sharing Consent View.

    Allows to request the DSC from a learner for a specific course.
    """
    template = 'enterprise/admin/clear_learners_data_sharing_consent.html'

    class ContextParameters:
        """
        Namespace-style class for custom context parameters.
        """
        ENTERPRISE_CUSTOMER = 'enterprise_customer'
        MANAGE_LEARNERS_DSC_FORM = 'manage_learners_data_sharing_consent_form'

    def get(self, request, customer_uuid):
        """
        Handle GET request - render "Request a DSC from Learner" form.

        Arguments:
            request (django.http.request.HttpRequest): Request instance
            customer_uuid (str): Enterprise Customer UUID

        Returns:
            django.http.response.HttpResponse: HttpResponse
        """
        return self.get_form_view(
            request,
            customer_uuid,
            additional_context={
                self.ContextParameters.MANAGE_LEARNERS_DSC_FORM: ManageLearnersDataSharingConsentForm()
            }
        )

    def post(self, request, customer_uuid):
        """
        Handle POST request - handle form submissions.

        Arguments:
            request (django.http.request.HttpRequest): Request instance
            customer_uuid (str): Enterprise Customer UUID
        """
        manage_learners_dsc_form = ManageLearnersDataSharingConsentForm(
            request.POST,
            enterprise_customer=EnterpriseCustomer.objects.get(uuid=customer_uuid)
        )

        # check that form data is well-formed
        if manage_learners_dsc_form.is_valid():
            delete_data_sharing_consent(
                manage_learners_dsc_form.cleaned_data[ManageLearnersDataSharingConsentForm.Fields.COURSE],
                customer_uuid,
                manage_learners_dsc_form.cleaned_data[ManageLearnersDataSharingConsentForm.Fields.EMAIL_OR_USERNAME]
            )
            messages.success(request, _("Successfully requested the Data Sharing consent from learner."))
            return HttpResponseRedirect(reverse("admin:" + UrlNames.MANAGE_LEARNERS_DSC, args=(customer_uuid,)))

        return self.get_form_view(
            request,
            customer_uuid,
            additional_context={
                self.ContextParameters.MANAGE_LEARNERS_DSC_FORM: manage_learners_dsc_form
            }
        )


class EnterpriseCustomerTransmitCoursesView(BaseEnterpriseCustomerView):
    """
    Transmit courses view.

    Allows transmitting of courses metadata for provided enterprise.
    """
    template = 'enterprise/admin/transmit_courses_metadata.html'

    class ContextParameters:
        """
        Namespace-style class for custom context parameters.
        """
        ENTERPRISE_CUSTOMER = 'enterprise_customer'
        TRANSMIT_COURSES_METADATA_FORM = 'transmit_courses_metadata_form'

    def get(self, request, customer_uuid):
        """
        Handle GET request - render "Transmit courses metadata" form.

        Arguments:
            request (django.http.request.HttpRequest): Request instance
            customer_uuid (str): Enterprise Customer UUID

        Returns:
            django.http.response.HttpResponse: HttpResponse
        """
        return self.get_form_view(
            request,
            customer_uuid,
            additional_context={
                self.ContextParameters.TRANSMIT_COURSES_METADATA_FORM: TransmitEnterpriseCoursesForm()
            }
        )

    def post(self, request, customer_uuid):
        """
        Handle POST request - handle form submissions.

        Arguments:
            request (django.http.request.HttpRequest): Request instance
            customer_uuid (str): Enterprise Customer UUID
        """
        transmit_courses_metadata_form = TransmitEnterpriseCoursesForm(request.POST)

        # check that form data is well-formed
        if transmit_courses_metadata_form.is_valid():
            channel_worker_username = transmit_courses_metadata_form.cleaned_data['channel_worker_username']

            # call `transmit_content_metadata` management command to trigger
            # transmission of enterprise courses metadata
            call_command(
                'transmit_content_metadata',
                '--catalog_user', channel_worker_username,
                enterprise_customer=customer_uuid
            )

            # Redirect to GET
            return HttpResponseRedirect('')

        return self.get_form_view(
            request,
            customer_uuid,
            additional_context={
                self.ContextParameters.TRANSMIT_COURSES_METADATA_FORM: transmit_courses_metadata_form
            }
        )


class EnterpriseCustomerManageLearnersView(BaseEnterpriseCustomerView):
    """
    Manage Learners view.

    Lists learners linked to chosen Enterprise Customer and allows adding and deleting them.
    """
    template = "enterprise/admin/manage_learners.html"

    class ContextParameters:
        """
        Namespace-style class for custom context parameters.
        """
        ENTERPRISE_CUSTOMER = "enterprise_customer"
        LEARNERS = "learners"
        PENDING_LEARNERS = "pending_learners"
        MANAGE_LEARNERS_FORM = "manage_learners_form"
        SEARCH_KEYWORD = "search_keyword"
        ENROLLMENT_URL = 'ENROLLMENT_API_ROOT_URL'

    def _get_view_context(self, request, customer_uuid):
        enterprise_customer = EnterpriseCustomer.objects.get(uuid=customer_uuid)
        search_keyword = self.get_search_keyword(request)
        linked_learners = self.get_enterprise_customer_user_queryset(request, search_keyword, customer_uuid)
        pending_linked_learners = self.get_pending_users_queryset(search_keyword, customer_uuid)

        context = {
            self.ContextParameters.ENTERPRISE_CUSTOMER: enterprise_customer,
            self.ContextParameters.PENDING_LEARNERS: pending_linked_learners,
            self.ContextParameters.LEARNERS: linked_learners,
            self.ContextParameters.SEARCH_KEYWORD: search_keyword or '',
            self.ContextParameters.ENROLLMENT_URL: settings.LMS_ENROLLMENT_API_PATH,
        }
        return context

    def get_search_keyword(self, request):
        """
        Retrieve the search querystring from the GET parameters.
        """
        return request.GET.get('q', None)

    def get_enterprise_customer_user_queryset(self, request, search_keyword, customer_uuid, page_size=PAGE_SIZE):
        """
        Get the list of EnterpriseCustomerUsers we want to render. Note that,
        if no ``search_keyword`` is provided, the resulting queryset will
        only include up to some constant limit of results, so that we
        can support searching for particular user records in customers
        with very large numbers of learners.

        Arguments:
            request (HttpRequest): HTTP Request instance.
            search_keyword (str): The keyword to search for in users' email addresses and usernames.
            customer_uuid (str): A unique identifier to filter down to only users linked to a
            particular EnterpriseCustomer.
            page_size (int): Number of learners displayed in each paginated set.
        """
        page = request.GET.get('page', 1)
        if search_keyword:
            enterprise_learners = self._get_searched_enterprise_customer_users(
                search_keyword,
                customer_uuid,
            )
        else:
            enterprise_learners = EnterpriseCustomerUser.objects.filter(
                enterprise_customer__uuid=customer_uuid,
            )[:DJANGO_ADMIN_MANAGE_LEARNERS_LIMIT]
        return paginated_list(enterprise_learners, page, page_size)

    def _get_searched_enterprise_customer_users(self, search_keyword, customer_uuid):
        """
        Helper to return a (possibly truncated) queryset of enterprise customer users
        who match some search criteria based on corresponding ``core.User`` records.
        """
        matching_user_ids = User.objects.filter(
            Q(email__icontains=search_keyword) | Q(username__icontains=search_keyword)
        ).values_list('id', flat=True)
        return EnterpriseCustomerUser.objects.filter(
            enterprise_customer__uuid=customer_uuid,
            user_id__in=matching_user_ids,
        )[:DJANGO_ADMIN_MANAGE_LEARNERS_LIMIT]

    def get_pending_users_queryset(self, search_keyword, customer_uuid):
        """
        Get the list of PendingEnterpriseCustomerUsers we want to render.

        Args:
            search_keyword (str): The keyword to search for in pending users' email addresses.
            customer_uuid (str): A unique identifier to filter down to only pending users
            linked to a particular EnterpriseCustomer.
        """
        if search_keyword:
            return PendingEnterpriseCustomerUser.objects.filter(
                enterprise_customer__uuid=customer_uuid,
                user_email__icontains=search_keyword,
            )[:DJANGO_ADMIN_MANAGE_LEARNERS_LIMIT]
        else:
            return PendingEnterpriseCustomerUser.objects.filter(
                enterprise_customer__uuid=customer_uuid
            )[:DJANGO_ADMIN_MANAGE_LEARNERS_LIMIT]

    @classmethod
    def _handle_singular(cls, request, enterprise_customer, manage_learners_form):
        """
        Link single user by email or username.

        Arguments:
            enterprise_customer (EnterpriseCustomer): learners will be linked to this Enterprise Customer instance
            manage_learners_form (ManageLearnersForm): bound ManageLearners form instance
        """
        form_field_value = manage_learners_form.cleaned_data[ManageLearnersForm.Fields.EMAIL_OR_USERNAME]
        email = email_or_username__to__email(form_field_value)
        try:
            validate_email_to_link(
                email,
                enterprise_customer,
                form_field_value,
                ValidationMessages.INVALID_EMAIL_OR_USERNAME,
                raise_exception=False,
            )
            EnterpriseCustomerUser.all_objects.link_user(enterprise_customer, email)
            return [email]
        except (ValidationError, LinkUserToEnterpriseError) as exc:
            manage_learners_form.add_error(ManageLearnersForm.Fields.EMAIL_OR_USERNAME, exc)
            return None

    def _handle_bulk_upload_errors(self, manage_learners_form, errors):
        """
        Handles adding errors to the ``manager_learners_form``.

        Arguments:
            manage_learners_form (ManageLearnersForm): bound ManageLearners form instance
            errors (list): List of validation errors from parsing the uploaded CSV.
        """
        manage_learners_form.add_error(
            ManageLearnersForm.Fields.GENERAL_ERRORS, ValidationMessages.BULK_LINK_FAILED
        )
        for error in errors:
            manage_learners_form.add_error(ManageLearnersForm.Fields.BULK_UPLOAD, error)

    def _process_bulk_upload_data(
            self,
            request,
            enterprise_customer,
            emails,
            already_linked_emails,
            duplicate_emails,
    ):
        """
        Processes the parsed data from bulk csv upload, adding the appropriate success and warning messages.

        Arguments:
            request: A request instance
            enterprise_customer (EnterpriseCustomer): An instance of an EnterpriseCustomer record.
            emails (list): List of valid, non-duplicate emails.
            already_linked_emails (list): List of emails that are already linked with the EnterpriseCustomer.
            duplicate_emails (list): List of emails that are duplicate entries.

        Returns: List of processable email addresses.
        """
        count = len(emails)
        messages.success(request, ngettext(
            "{count} new learner was added to {enterprise_customer_name}.",
            "{count} new learners were added to {enterprise_customer_name}.",
            count
        ).format(count=count, enterprise_customer_name=enterprise_customer.name))
        this_customer_linked_emails = [
            email for email, customer in already_linked_emails if customer == enterprise_customer
        ]
        other_customer_linked_emails = [
            email for email, __ in already_linked_emails if email not in this_customer_linked_emails
        ]
        if this_customer_linked_emails:
            messages.warning(
                request,
                _(
                    "The following learners were already associated with this Enterprise "
                    "Customer: {list_of_emails}"
                ).format(
                    list_of_emails=", ".join(this_customer_linked_emails)
                )
            )
        if other_customer_linked_emails:
            messages.warning(
                request,
                _(
                    "The following learners are already associated with "
                    "another Enterprise Customer. These learners were not "
                    "added to {enterprise_customer_name}: {list_of_emails}"
                ).format(
                    enterprise_customer_name=enterprise_customer.name,
                    list_of_emails=", ".join(other_customer_linked_emails),
                )
            )
        if duplicate_emails:
            messages.warning(
                request,
                _(
                    "The following duplicate email addresses were not added: "
                    "{list_of_emails}"
                ).format(
                    list_of_emails=", ".join(duplicate_emails)
                )
            )
        # Build a list of all the emails that we can act on further; that is,
        # emails that we either linked to this customer, or that were linked already.
        all_processable_emails = list(emails) + this_customer_linked_emails
        return all_processable_emails

    @classmethod
    def _handle_bulk_upload(cls, enterprise_customer, manage_learners_form, request, email_list=None):
        """
        Bulk link users by email.

        Arguments:
            enterprise_customer (EnterpriseCustomer): learners will be linked to this Enterprise Customer instance
            manage_learners_form (ManageLearnersForm): bound ManageLearners form instance
            request (django.http.request.HttpRequest): HTTP Request instance
            email_list (iterable): A list of pre-processed email addresses to handle using the form
        """
        errors = []
        emails = set()
        course_id_with_emails = {}
        already_linked_emails = []
        duplicate_emails = []
        csv_file = manage_learners_form.cleaned_data[ManageLearnersForm.Fields.BULK_UPLOAD]
        if email_list:
            parsed_csv = [{ManageLearnersForm.CsvColumns.EMAIL: email} for email in email_list]
        else:
            parsed_csv = parse_csv(csv_file, expected_columns={ManageLearnersForm.CsvColumns.EMAIL})

        try:
            for index, row in enumerate(parsed_csv):
                email = row[ManageLearnersForm.CsvColumns.EMAIL]
                course_id = row.get(ManageLearnersForm.CsvColumns.COURSE_ID, None)  # optional column
                course_details = None
                try:
                    already_linked = validate_email_to_link(email, enterprise_customer, raise_exception=False)
                    if course_id:
                        course_details = validate_course_exists_for_enterprise(enterprise_customer, course_id)
                except ValidationError as exc:
                    message = _("Error at line {line}: {message}\n").format(line=index + 1, message=exc)
                    errors.append(message)
                else:
                    if already_linked:
                        already_linked_emails.append((email, already_linked.enterprise_customer))
                    elif email in emails:
                        duplicate_emails.append(email)
                    else:
                        emails.add(email)

                    # course column exists for row, is a valid course id, and exists in the enterprise's catalog(s).
                    if course_details:
                        if course_details['course_id'] not in course_id_with_emails:
                            course_id_with_emails[course_details['course_id']] = {email}
                        else:
                            course_id_with_emails[course_details['course_id']].add(email)
        except ValidationError as exc:
            errors.append(exc)

        # do the actual linking if there are no errors at this point:
        if not errors:
            try:
                with transaction.atomic():
                    for email in emails:
                        EnterpriseCustomerUser.all_objects.link_user(enterprise_customer, email)
            except LinkUserToEnterpriseError as exc:
                errors.append(exc)

        if errors:
            cls._handle_bulk_upload_errors(cls, manage_learners_form=manage_learners_form, errors=errors)
            # There were validation errors, so prevent any further action.
            return [], {}

        # Process the bulk uploaded data:
        processable_emails = cls._process_bulk_upload_data(
            cls,
            request=request,
            enterprise_customer=enterprise_customer,
            emails=emails,
            already_linked_emails=already_linked_emails,
            duplicate_emails=duplicate_emails,
        )
        return processable_emails, course_id_with_emails

    @classmethod
    def send_messages(cls, http_request, message_requests):
        """
        Deduplicate any outgoing message requests, and send the remainder.

        Args:
            http_request: The HTTP request in whose response we want to embed the messages
            message_requests: A list of undeduplicated messages in the form of tuples of message type
                and text- for example, ('error', 'Something went wrong')
        """
        deduplicated_messages = set(message_requests)
        for msg_type, text in deduplicated_messages:
            message_function = getattr(messages, msg_type)
            message_function(http_request, text)

    @classmethod
    def get_success_enrollment_message(cls, users, enrolled_in):
        """
        Create message for the users who were enrolled in a course.

        Args:
            users: An iterable of users who were successfully enrolled
            enrolled_in (str): A string identifier for the course the users were enrolled in

        Returns:
            tuple: A 2-tuple containing a message type and message text
        """
        enrolled_count = len(users)
        return (
            'success',
            ngettext(
                '{enrolled_count} learner was enrolled in {enrolled_in}.',
                '{enrolled_count} learners were enrolled in {enrolled_in}.',
                enrolled_count,
            ).format(
                enrolled_count=enrolled_count,
                enrolled_in=enrolled_in,
            )
        )

    @classmethod
    def get_failed_enrollment_message(cls, users, enrolled_in):
        """
        Create message for the users who were not able to be enrolled in a course.

        Args:
            users: An iterable of users who were not successfully enrolled
            enrolled_in (str): A string identifier for the course with which enrollment was attempted

        Returns:
        tuple: A 2-tuple containing a message type and message text
        """
        failed_emails = [user.email for user in users]
        return (
            'error',
            _(
                'The following learners could not be enrolled in {enrolled_in}: {user_list}'
            ).format(
                enrolled_in=enrolled_in,
                user_list=', '.join(failed_emails),
            )
        )

    @classmethod
    def get_pending_enrollment_message(cls, pending_users, enrolled_in):
        """
        Create message for the users who were enrolled in a course.

        Args:
            users: An iterable of PendingEnterpriseCustomerUsers who were successfully linked with a pending enrollment
            enrolled_in (str): A string identifier for the course the pending users were linked to

        Returns:
            tuple: A 2-tuple containing a message type and message text
        """
        pending_emails = [pending_user.user_email for pending_user in pending_users]
        return (
            'warning',
            _(
                "The following learners do not have an account on "
                "{platform_name}. They have not been enrolled in "
                "{enrolled_in}. When these learners create an account, they will "
                "be enrolled automatically: {pending_email_list}"
            ).format(
                platform_name=settings.PLATFORM_NAME,
                enrolled_in=enrolled_in,
                pending_email_list=', '.join(pending_emails),
            )
        )

    @classmethod
    def _enroll_users(
            cls,
            request,
            enterprise_customer,
            emails,
            mode,
            course_id=None,
            notify=True,
            enrollment_reason=None,
            sales_force_id=None,
            discount=0.0
    ):
        """
        Enroll the users with the given email addresses to the course.

        Args:
            cls (type): The EnterpriseCustomerManageLearnersView class itself
            request: The HTTP request the enrollment is being created by
            enterprise_customer: The instance of EnterpriseCustomer whose attached users we're enrolling
            emails: An iterable of strings containing email addresses to enroll in a course
            mode: The enrollment mode the users will be enrolled in the course with
            course_id: The ID of the course in which we want to enroll
            notify: Whether to notify (by email) the users that have been enrolled
        """
        pending_messages = []
        paid_modes = ['verified', 'professional']

        succeeded, pending, failed = enroll_users_in_course(
            enterprise_customer=enterprise_customer,
            course_id=course_id,
            course_mode=mode,
            emails=sorted(emails),
            enrollment_requester=request.user,
            enrollment_reason=enrollment_reason,
            discount=discount,
            sales_force_id=sales_force_id,
        )
        all_successes = succeeded + pending
        if notify:
            enterprise_customer.notify_enrolled_learners(
                catalog_api_user=request.user,
                course_id=course_id,
                users=all_successes,
            )
        if succeeded:
            pending_messages.append(cls.get_success_enrollment_message(succeeded, course_id))
        if failed:
            pending_messages.append(cls.get_failed_enrollment_message(failed, course_id))
        if pending:
            pending_messages.append(cls.get_pending_enrollment_message(pending, course_id))

        if mode in paid_modes:
            # Create an order to track the manual enrollments of non-pending accounts
            enrollments = [{
                "lms_user_id": success.id,
                "email": success.email,
                "username": success.username,
                "course_run_key": course_id,
                "discount_percentage": float(discount),
                "enterprise_customer_name": enterprise_customer.name,
                "enterprise_customer_uuid": str(enterprise_customer.uuid),
                "mode": mode,
                "sales_force_id": sales_force_id,
            } for success in succeeded]
            EcommerceApiClient(get_ecommerce_worker_user()).create_manual_enrollment_orders(enrollments)
        cls.send_messages(request, pending_messages)

    def get(self, request, customer_uuid):
        """
        Handle GET request - render linked learners list and "Link learner" form.

        Arguments:
            request (django.http.request.HttpRequest): Request instance
            customer_uuid (str): Enterprise Customer UUID

        Returns:
            django.http.response.HttpResponse: HttpResponse
        """
        enterprise_customer = EnterpriseCustomer.objects.get(uuid=customer_uuid)
        manage_learners_form = ManageLearnersForm(user=request.user, enterprise_customer=enterprise_customer)

        return self.get_form_view(
            request,
            customer_uuid,
            additional_context={
                self.ContextParameters.MANAGE_LEARNERS_FORM: manage_learners_form
            }
        )

    def post(self, request, customer_uuid):
        """
        Handle POST request - handle form submissions.

        Arguments:
            request (django.http.request.HttpRequest): Request instance
            customer_uuid (str): Enterprise Customer UUID

        Returns:
            django.http.response.HttpResponse: HttpResponse
        """
        enterprise_customer = EnterpriseCustomer.objects.get(uuid=customer_uuid)
        manage_learners_form = ManageLearnersForm(
            request.POST,
            request.FILES,
            user=request.user,
            enterprise_customer=enterprise_customer
        )
        course_id_with_emails = {}

        # initial form validation - check that form data is well-formed
        if manage_learners_form.is_valid():
            email_field_as_bulk_input = split_usernames_and_emails(
                manage_learners_form.cleaned_data[ManageLearnersForm.Fields.EMAIL_OR_USERNAME]
            )
            is_bulk_entry = len(email_field_as_bulk_input) > 1
            # The form is valid. Call the appropriate helper depending on the mode:
            mode = manage_learners_form.cleaned_data[ManageLearnersForm.Fields.MODE]
            if mode == ManageLearnersForm.Modes.MODE_SINGULAR and not is_bulk_entry:
                linked_learners = self._handle_singular(request, enterprise_customer, manage_learners_form)
            elif mode == ManageLearnersForm.Modes.MODE_SINGULAR:
                linked_learners, __ = self._handle_bulk_upload(
                    enterprise_customer,
                    manage_learners_form,
                    request,
                    email_list=email_field_as_bulk_input
                )
            else:
                linked_learners, course_id_with_emails = self._handle_bulk_upload(
                    enterprise_customer,
                    manage_learners_form,
                    request,
                )

        # _handle_form might add form errors, so we check if it is still valid
        if manage_learners_form.is_valid():
            # If we aren't installed in Open edX, blank out enrollment reason so downstream methods don't attempt to
            # create audit items
            if create_manual_enrollment_audit is not None:
                manual_enrollment_reason = manage_learners_form.cleaned_data.get(ManageLearnersForm.Fields.REASON)
            else:
                manual_enrollment_reason = None
                LOG.exception(
                    "To create enrollment audits for enterprise learners, "
                    "this package must be installed in an Open edX environment."
                )

            notification_type = manage_learners_form.cleaned_data.get(ManageLearnersForm.Fields.NOTIFY)
            notify = notification_type == ManageLearnersForm.NotificationTypes.BY_EMAIL
            discount = manage_learners_form.cleaned_data.get(ManageLearnersForm.Fields.DISCOUNT)
            sales_force_id = manage_learners_form.cleaned_data.get(ManageLearnersForm.Fields.SALES_FORCE_ID)
            course_mode = manage_learners_form.cleaned_data.get(ManageLearnersForm.Fields.COURSE_MODE)
            course_id = None

            if not course_id_with_emails:
                course_details = manage_learners_form.cleaned_data.get(ManageLearnersForm.Fields.COURSE) or {}
                course_id = course_details.get('course_id')
                if course_id and linked_learners:
                    self._enroll_users(
                        request=request,
                        enterprise_customer=enterprise_customer,
                        emails=linked_learners,
                        mode=course_mode,
                        course_id=course_id,
                        notify=notify,
                        enrollment_reason=manual_enrollment_reason,
                        sales_force_id=sales_force_id,
                        discount=discount
                    )
            else:
                for course_id, emails in course_id_with_emails.items():
                    emails_to_enroll = emails
                    if emails_to_enroll:
                        self._enroll_users(
                            request=request,
                            enterprise_customer=enterprise_customer,
                            emails=list(emails_to_enroll),
                            mode=course_mode,
                            course_id=course_id,
                            notify=notify,
                            enrollment_reason=manual_enrollment_reason,
                            sales_force_id=sales_force_id,
                            discount=discount
                        )

            # Redirect to GET if everything went smooth.
            manage_learners_url = reverse("admin:" + UrlNames.MANAGE_LEARNERS, args=(customer_uuid,))
            search_keyword = self.get_search_keyword(request)
            if search_keyword:
                manage_learners_url = manage_learners_url + "?q=" + search_keyword
            return HttpResponseRedirect(manage_learners_url)

        # if something went wrong - display bound form on the page
        return self.get_form_view(
            request,
            customer_uuid,
            additional_context={
                self.ContextParameters.MANAGE_LEARNERS_FORM: manage_learners_form
            }
        )

    def delete(self, request, customer_uuid):
        """
        Handle DELETE request - handle unlinking learner.

        Arguments:
            request (django.http.request.HttpRequest): Request instance
            customer_uuid (str): Enterprise Customer UUID

        Returns:
            django.http.response.HttpResponse: HttpResponse
        """
        # TODO: pylint acts stupid - find a way around it without suppressing
        enterprise_customer = EnterpriseCustomer.objects.get(uuid=customer_uuid)
        email_to_unlink = request.GET["unlink_email"]
        try:
            EnterpriseCustomerUser.objects.unlink_user(
                enterprise_customer=enterprise_customer,
                user_email=email_to_unlink,
            )
        except (EnterpriseCustomerUser.DoesNotExist, PendingEnterpriseCustomerUser.DoesNotExist):
            message = _("Email {email} is not associated with Enterprise "
                        "Customer {ec_name}").format(
                            email=email_to_unlink, ec_name=enterprise_customer.name)
            # pylint: disable=http-response-with-content-type-json
            return HttpResponse(message, content_type="application/json", status=404)

        return JsonResponse({})


class EnterpriseCustomerSetupAuthOrgIDView(BaseEnterpriseCustomerView):
    """
    Setup Auth org id View.

    This action will configure SSO to GetSmarter using edX credentials via Auth0.
    """
    template = 'enterprise/admin/setup_auth_org_id.html'

    def get(self, request, customer_uuid):
        """
        Handle GET request - render "Setup Auth org id" form.

        Arguments:
            request (django.http.request.HttpRequest): Request instance
            customer_uuid (str): Enterprise Customer UUID

        Returns:
            django.http.response.HttpResponse: HttpResponse
        """
        if not get_sso_orchestrator_configure_edx_oauth_path():
            return HttpResponseForbidden(
                "The ENTERPRISE_SSO_ORCHESTRATOR_CONFIGURE_EDX_OAUTH_PATH setting was not configured."
            )
        return self.get_form_view(request, customer_uuid)

    def post(self, request, customer_uuid):
        """
        Handle POST request - handle form submissions.

        Arguments:
            request (django.http.request.HttpRequest): Request instance
            customer_uuid (str): Enterprise Customer UUID
        """
        if not get_sso_orchestrator_configure_edx_oauth_path():
            return HttpResponseForbidden(
                "The ENTERPRISE_SSO_ORCHESTRATOR_CONFIGURE_EDX_OAUTH_PATH setting was not configured."
            )

        enterprise_customer = EnterpriseCustomer.objects.get(uuid=customer_uuid)

        # Call the configure-edx-oauth endpoint on the enterprise-sso-orchestrator service to obtain an orgId.
        # This will raise SsoOrchestratorClientError if the API request fails.
        try:
            auth_org_id = EnterpriseSSOOrchestratorApiClient().configure_edx_oauth(enterprise_customer)
        except SsoOrchestratorClientError as exc:
            error_msg = (
                f"Error configuring edx oauth for enterprise customer {enterprise_customer.name}"
                f"<{enterprise_customer.uuid}>: {exc}"
            )
            LOG.exception(error_msg)
            return HttpResponseServerError(error_msg)

        if auth_org_id:
            enterprise_customer.auth_org_id = auth_org_id
            enterprise_customer.save()
            messages.success(request, _('Successfully written the "Auth org id" field for this enterprise customer.'))
            return HttpResponseRedirect(reverse("admin:" + UrlNames.SETUP_AUTH_ORG_ID, args=(customer_uuid,)))
        else:
            # Annoyingly, there's still the remote possibility that the request succeeded but we failed to retrieve the
            # auth_org_id. This might be due to a regression in the API response schema.
            error_msg = (
                f"Error configuring edx oauth for enterprise customer {enterprise_customer.name}"
                f"<{enterprise_customer.uuid}>: Missing orgId."
            )
            LOG.exception(error_msg)
            return HttpResponseServerError(error_msg)



"""
Python API for various enterprise functionality.
"""
from enterprise import roles_api
from enterprise.models import EnterpriseCustomerAdmin, PendingEnterpriseCustomerAdminUser


def activate_admin_permissions(enterprise_customer_user):
    """
    Activates admin permissions for an existing PendingEnterpriseCustomerAdminUser.

    Specifically, the "enterprise_admin" system-wide role is assigned to the user and
    the PendingEnterpriseCustomerAdminUser record is removed.

    Requires an EnterpriseCustomerUser record to exist which ensures the user already
    has the "enterprise_learner" role as a prerequisite.

    Arguments:
        enterprise_customer_user: an EnterpriseCustomerUser instance
    """
    try:
        pending_admin_user = PendingEnterpriseCustomerAdminUser.objects.get(
            user_email=enterprise_customer_user.user.email,
            enterprise_customer=enterprise_customer_user.enterprise_customer,
        )
        # if this user is an admin, we want to create an accompanying EnterpriseCustomerAdmin record
        EnterpriseCustomerAdmin.objects.get_or_create(
            enterprise_customer_user=enterprise_customer_user,
        )
    except PendingEnterpriseCustomerAdminUser.DoesNotExist:
        return  # this is ok, nothing to do

    if not enterprise_customer_user.linked:
        # EnterpriseCustomerUser is no longer linked, so delete the "enterprise_admin" role.
        # TODO: ENT-3914 | Add `enterprise_customer=enterprise_customer_user.enterprise_customer`
        # kwarg so that we delete at most a single assignment instance.
        roles_api.delete_admin_role_assignment(
            enterprise_customer_user.user,
        )
        return  # nothing left to do

    roles_api.assign_admin_role(
        enterprise_customer_user.user,
        enterprise_customer=enterprise_customer_user.enterprise_customer
    )

    # delete the PendingEnterpriseCustomerAdminUser record
    pending_admin_user.delete()



"""
Django management command to validate that DefaultEnterpriseEnrollmentIntention
objects have enrollable content.
"""
import logging
from datetime import timedelta

from django.core.management import BaseCommand, CommandError
from django.db.models import DateTimeField, Max
from django.db.models.functions import Cast, Coalesce, Greatest
from django.utils import timezone

from enterprise.content_metadata.api import get_and_cache_customer_content_metadata
from enterprise.models import DefaultEnterpriseEnrollmentIntention

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    """
    Enumerate the catalog filters and log information about how we might migrate them.
    """

    def __init__(self, *args, **kwargs):
        self.delay_minutes = None
        super().__init__(*args, **kwargs)

    def add_arguments(self, parser):
        parser.add_argument(
            '--delay-minutes',
            dest='delay_minutes',
            required=False,
            type=int,
            default=30,
            help="How long after a customer's catalog has been updated are we allowed to evaluate the customer."
        )

    @property
    def latest_change_allowed(self):
        return timezone.now() - timedelta(minutes=self.delay_minutes)

    def handle_intention(self, intention):
        """
        Check that the default enrollment intention's content_key is contained in any of the customer's catalogs.

        Returns:
            dict: Results dict that indicates whether evaluation was skipped, and whether the intention was valid.
        """
        customer = intention.enterprise_customer
        result = {
            'skipped': None,
            'invalid': None,
        }

        if not intention.catalogs_modified_latest:
            result['skipped'] = True
            logger.info(
                f"handle_intention(): SKIPPING Evaluating enrollment intention {intention} "
                "for not having any related catalogs."
            )
            return result
        if intention.catalogs_modified_latest > self.latest_change_allowed:
            result['skipped'] = True
            logger.info(
                f"handle_intention(): SKIPPING Evaluating enrollment intention {intention} "
                "for having catalogs which have been too recently updated."
            )
            return result
        result['skipped'] = False
        logger.info(f"handle_intention(): Evaluating enrollment intention {intention}.")

        content_metadata = get_and_cache_customer_content_metadata(
            customer.uuid,
            intention.content_key,
        )
        contained_in_customer_catalogs = bool(content_metadata)
        if contained_in_customer_catalogs:
            logger.info(
                f"handle_default_enrollment_intention(): Default enrollment intention {intention} "
                "is compatible with the customer's catalogs."
            )
            result["invalid"] = False
        else:
            logger.error(
                f"handle_default_enrollment_intention(): Default enrollment intention {intention} "
                "is NOT compatible with the customer's catalogs."
            )
            result["invalid"] = True
        return result

    def handle(self, *args, **options):
        self.delay_minutes = options.get("delay_minutes")

        intentions = DefaultEnterpriseEnrollmentIntention.objects.select_related(
            'enterprise_customer'
        ).prefetch_related(
            'enterprise_customer__enterprise_customer_catalogs'
        ).annotate(
            catalogs_modified_latest=Greatest(
                Max("enterprise_customer__enterprise_customer_catalogs__modified"),
                Coalesce(
                    Max("enterprise_customer__enterprise_customer_catalogs__enterprise_catalog_query__modified"),
                    # Note 1: Arbitrarily fallback to 1 year ago because Greatest() in MySQL relies on non-null inputs.
                    # Note 2: Cast python datetime to django field type, or else experience weird errors in prod.
                    Cast(timezone.now() - timedelta(days=360), DateTimeField())
                )
            )
        )

        results = {intention: self.handle_intention(intention) for intention in intentions}
        results_evaluated = {intention: result for intention, result in results.items() if not result['skipped']}
        results_invalid = {intention: result for intention, result in results_evaluated.items() if result['invalid']}

        count_total = len(results)
        count_evaluated = len(results_evaluated)
        count_skipped = count_total - count_evaluated
        count_invalid = len(results_invalid)
        count_passed = count_evaluated - count_invalid

        invalid_intentions = results_invalid.keys()

        logger.info(
            f"{count_total} total enrollment intentions found, "
            f"and {count_evaluated}/{count_total} were evaluated "
            f"({count_skipped}/{count_total} skipped)."
        )
        logger.info(
            f"Out of {count_evaluated} total evaluated enrollment intentions, "
            f"{count_passed}/{count_evaluated} passed validation "
            f"({count_invalid}/{count_evaluated} invalid)."
        )
        if count_invalid > 0:
            logger.error(f"Summary of all {count_invalid} invalid intentions: {invalid_intentions}")
            logger.error("FAILURE: Some default enrollment intentions were invalid.")
            raise CommandError(f"{count_invalid} invalid default enrollment intentions found.")
        logger.info("SUCCESS: All default enrollment intentions are valid!")



"""
enterprise_learner_portal utils.
"""
from datetime import datetime

from pytz import utc


class CourseRunProgressStatuses:
    """
    Class to group statuses that a course run can be in with respect to user progress.
    """

    IN_PROGRESS = 'in_progress'
    UPCOMING = 'upcoming'
    COMPLETED = 'completed'
    SAVED_FOR_LATER = 'saved_for_later'


def get_course_run_status(course_overview, certificate_info, enterprise_enrollment):
    """
    Get the status of a course run, given the state of a user's certificate in the course.

    A run is considered "complete" when either the course run has ended OR the user has earned a
    passing certificate.

    Arguments:
        course_overview (CourseOverview): the overview for the course run
        certificate_info: A dict containing the following key:
            ``is_passing``: whether the  user has a passing certificate in the course run

    Returns:
        status: one of (
            CourseRunProgressStatuses.SAVED_FOR_LATER,
            CourseRunProgressStatuses.COMPLETE,
            CourseRunProgressStatuses.IN_PROGRESS,
            CourseRunProgressStatuses.UPCOMING,
        )
    """
    if enterprise_enrollment and enterprise_enrollment.saved_for_later:
        return CourseRunProgressStatuses.SAVED_FOR_LATER

    is_certificate_passing = certificate_info.get('is_passing', False)

    if course_overview['has_ended'] or is_certificate_passing:
        return CourseRunProgressStatuses.COMPLETED
    if course_overview['has_started']:
        return CourseRunProgressStatuses.IN_PROGRESS
    return CourseRunProgressStatuses.UPCOMING


def get_exec_ed_course_run_status(course_details, certificate_info, enterprise_enrollment):
    """
    Get the status of a exec ed course run, given the state of a user's certificate in the course.

    A run is considered "complete" when either the course run has ended OR the user has earned a
    passing certificate.

    Arguments:
        course_details : the details for the exececutive education course run
        certificate_info: A dict containing the following key:
            ``is_passing``: whether the  user has a passing certificate in the course run

    Returns:
        status: one of (
            CourseRunProgressStatuses.SAVED_FOR_LATER,
            CourseRunProgressStatuses.COMPLETE,
            CourseRunProgressStatuses.IN_PROGRESS,
            CourseRunProgressStatuses.UPCOMING,
        )
    """
    if enterprise_enrollment and enterprise_enrollment.saved_for_later:
        return CourseRunProgressStatuses.SAVED_FOR_LATER

    is_certificate_passing = certificate_info.get('is_passing', False)
    start_date = course_details.start_date
    end_date = course_details.end_date

    has_started = datetime.now(utc) > start_date if start_date is not None else True
    has_ended = datetime.now(utc) > end_date if end_date is not None else False

    if has_ended or is_certificate_passing:
        return CourseRunProgressStatuses.COMPLETED
    if has_started:
        return CourseRunProgressStatuses.IN_PROGRESS
    return CourseRunProgressStatuses.UPCOMING



# Generated by Django 4.2.13 on 2024-06-12 07:54

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('enterprise', '0210_enterprisecustomer_enable_learner_portal_sidebar_message_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='enterprisecustomer',
            name='show_videos_in_learner_portal_search_results',
            field=models.BooleanField(default=False, help_text='If checked, videos will be displayed in the search results on the learner portal.', verbose_name='Show videos in learner portal search results'),
        ),
        migrations.AddField(
            model_name='historicalenterprisecustomer',
            name='show_videos_in_learner_portal_search_results',
            field=models.BooleanField(default=False, help_text='If checked, videos will be displayed in the search results on the learner portal.', verbose_name='Show videos in learner portal search results'),
        ),
    ]



# Generated by Django 3.2.20 on 2023-08-08 09:23

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('cornerstone', '0028_auto_20230719_1621'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='historicalcornerstoneenterprisecustomerconfiguration',
            options={'get_latest_by': ('history_date', 'history_id'), 'ordering': ('-history_date', '-history_id'), 'verbose_name': 'historical cornerstone enterprise customer configuration', 'verbose_name_plural': 'historical cornerstone enterprise customer configurations'},
        ),
    ]



"""
Database models for Enterprise Integrated Channel.
"""

import json
import logging

from jsonfield.fields import JSONField

from django.contrib import auth
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import Q
from django.db.models.query import QuerySet
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from model_utils.models import TimeStampedModel

from enterprise.constants import TRANSMISSION_MARK_CREATE, TRANSMISSION_MARK_DELETE, TRANSMISSION_MARK_UPDATE
from enterprise.models import EnterpriseCustomer, EnterpriseCustomerCatalog
from enterprise.utils import localized_utcnow
from integrated_channels.integrated_channel.exporters.content_metadata import ContentMetadataExporter
from integrated_channels.integrated_channel.exporters.learner_data import LearnerExporter
from integrated_channels.integrated_channel.transmitters.content_metadata import ContentMetadataTransmitter
from integrated_channels.integrated_channel.transmitters.learner_data import LearnerTransmitter
from integrated_channels.utils import channel_code_to_app_label, convert_comma_separated_string_to_list

LOGGER = logging.getLogger(__name__)
User = auth.get_user_model()
LAST_24_HRS = timezone.now() - timezone.timedelta(hours=24)


def set_default_display_name(*args, **kw):
    """
    post_save signal reciever to set default display name
    wired up in EnterpriseCustomerPluginConfiguration.__init_subclass__
    """
    this_display_name = kw['instance'].display_name
    # check if display_name is None, empty, or just spaces
    if not (this_display_name and this_display_name.strip()):
        kw['instance'].display_name = kw['instance'].generate_default_display_name()
        kw['instance'].save()


class SoftDeletionQuerySet(QuerySet):
    """
    Soft deletion query set.
    """

    def delete(self):
        return super().update(deleted_at=localized_utcnow())

    def hard_delete(self):
        return super().delete()

    def revive(self):
        return super().update(deleted_at=None)


class SoftDeletionManager(models.Manager):
    """
    Soft deletion manager overriding a model's query set in order to soft delete.
    """
    use_for_related_fields = True

    def __init__(self, *args, **kwargs):
        self.alive_only = kwargs.pop('alive_only', True)
        super().__init__(*args, **kwargs)

    def get_queryset(self):
        if self.alive_only:
            return SoftDeletionQuerySet(self.model, using=self._db, hints=self._hints).filter(deleted_at=None)
        return SoftDeletionQuerySet(self.model)

    def delete(self):
        return self.get_queryset().delete()

    def hard_delete(self):
        return self.get_queryset().hard_delete()

    def revive(self):
        return self.get_queryset().revive()


class SoftDeletionModel(TimeStampedModel):
    """
    Soft deletion model that sets a particular entries `deleted_at` field instead of removing the entry on delete.
    Use `hard_delete()` to permanently remove entries.
    """
    deleted_at = models.DateTimeField(blank=True, null=True)

    objects = SoftDeletionManager()
    all_objects = SoftDeletionManager(alive_only=False)

    class Meta:
        abstract = True


class EnterpriseCustomerPluginConfiguration(SoftDeletionModel):
    """
    Abstract base class for information related to integrating with external systems for an enterprise customer.

    EnterpriseCustomerPluginConfiguration should be extended by configuration models in other integrated channel
    apps to provide uniformity across different integrated channels.

    The configuration provides default exporters and transmitters if the ``get_x_data_y`` methods aren't
    overridden, where ``x`` and ``y`` are (learner, course) and (exporter, transmitter) respectively.
    """

    display_name = models.CharField(
        max_length=255,
        blank=True,
        default='',
        help_text=_("A configuration nickname.")
    )

    enterprise_customer = models.ForeignKey(
        EnterpriseCustomer,
        blank=False,
        null=False,
        help_text=_("Enterprise Customer associated with the configuration."),
        on_delete=models.deletion.CASCADE
    )

    idp_id = models.CharField(
        max_length=255,
        blank=True,
        default='',
        help_text=_("If provided, will be used as IDP slug to locate remote id for learners")
    )

    active = models.BooleanField(
        blank=False,
        null=False,
        help_text=_("Is this configuration active?"),
    )

    dry_run_mode_enabled = models.BooleanField(
        blank=False,
        null=False,
        default=False,
        help_text=_("Is this configuration in dry-run mode? (experimental)"),
    )

    show_course_price = models.BooleanField(
        blank=False,
        null=False,
        default=False,
        help_text=_("Displays course price"),
    )

    transmission_chunk_size = models.IntegerField(
        default=500,
        help_text=_("The maximum number of data items to transmit to the integrated channel with each request.")
    )

    channel_worker_username = models.CharField(
        max_length=255,
        blank=True,
        default='',
        help_text=_("Enterprise channel worker username to get JWT tokens for authenticating LMS APIs."),
    )
    catalogs_to_transmit = models.TextField(
        blank=True,
        default='',
        help_text=_(
            "A comma-separated list of catalog UUIDs to transmit. If blank, all customer catalogs will be transmitted. "
            "If there are overlapping courses in the customer catalogs, the overlapping course metadata will be "
            "selected from the newest catalog."
        ),
    )
    disable_learner_data_transmissions = models.BooleanField(
        default=False,
        verbose_name="Disable Learner Data Transmission",
        help_text=_("When set to True, the configured customer will no longer receive learner data transmissions, both"
                    " scheduled and signal based")
    )
    last_sync_attempted_at = models.DateTimeField(
        help_text='The DateTime of the most recent Content or Learner data record sync attempt',
        blank=True,
        null=True
    )
    last_content_sync_attempted_at = models.DateTimeField(
        help_text='The DateTime of the most recent Content data record sync attempt',
        blank=True,
        null=True
    )
    last_learner_sync_attempted_at = models.DateTimeField(
        help_text='The DateTime of the most recent Learner data record sync attempt',
        blank=True,
        null=True
    )
    last_sync_errored_at = models.DateTimeField(
        help_text='The DateTime of the most recent failure of a Content or Learner data record sync attempt',
        blank=True,
        null=True
    )
    last_content_sync_errored_at = models.DateTimeField(
        help_text='The DateTime of the most recent failure of a Content data record sync attempt',
        blank=True,
        null=True
    )
    last_learner_sync_errored_at = models.DateTimeField(
        help_text='The DateTime of the most recent failure of a Learner data record sync attempt',
        blank=True,
        null=True
    )
    last_modified_at = models.DateTimeField(
        help_text='The DateTime of the last change made to this configuration.',
        auto_now=True,
        blank=True,
        null=True
    )

    class Meta:
        abstract = True

    @classmethod
    def __init_subclass__(cls, **kwargs):
        """
        Finds every subclass and wires up the signal reciever to set default display name when blank
        """
        super().__init_subclass__(**kwargs)
        models.signals.post_save.connect(set_default_display_name, sender=cls)

    def delete(self, *args, **kwargs):
        self.deleted_at = localized_utcnow()
        self.save()

    def revive(self):
        self.deleted_at = None
        self.save()

    def hard_delete(self):
        return super().delete()

    def clean(self):
        invalid_uuids = []
        for uuid in convert_comma_separated_string_to_list(self.catalogs_to_transmit):
            try:
                EnterpriseCustomerCatalog.objects.get(uuid=uuid, enterprise_customer=self.enterprise_customer)
            except (EnterpriseCustomerCatalog.DoesNotExist, ValidationError):
                invalid_uuids.append(str(uuid))
        if invalid_uuids:
            raise ValidationError(
                {
                    'catalogs_to_transmit': [
                        "These are the invalid uuids: {invalid_uuids}".format(invalid_uuids=invalid_uuids)
                    ]
                }
            )

    def fetch_orphaned_content_audits(self):
        """
        Helper method attached to customer configs to fetch all orphaned content metadata audits not linked to the
        customer's catalogs.
        """
        enterprise_customer_catalogs = self.customer_catalogs_to_transmit or \
            self.enterprise_customer.enterprise_customer_catalogs.all()

        customer_catalog_uuids = enterprise_customer_catalogs.values_list('uuid', flat=True)

        non_existent_catalogs_filter = Q(enterprise_customer_catalog_uuid__in=customer_catalog_uuids)
        null_catalogs_filter = Q(enterprise_customer_catalog_uuid__isnull=True)

        return ContentMetadataItemTransmission.objects.filter(
            integrated_channel_code=self.channel_code(),
            enterprise_customer=self.enterprise_customer,
            remote_deleted_at__isnull=True,
            remote_created_at__isnull=False,
        ).filter(~non_existent_catalogs_filter | null_catalogs_filter)

    def update_content_synced_at(self, action_happened_at, was_successful):
        """
        Given the last time a Content record sync was attempted and status update the appropriate timestamps.
        """
        update_fields = []
        if self.last_sync_attempted_at is None or action_happened_at > self.last_sync_attempted_at:
            self.last_sync_attempted_at = action_happened_at
            update_fields.append('last_sync_attempted_at')
        if self.last_content_sync_attempted_at is None or action_happened_at > self.last_content_sync_attempted_at:
            self.last_content_sync_attempted_at = action_happened_at
            update_fields.append('last_content_sync_attempted_at')
        if not was_successful:
            if self.last_sync_errored_at is None or action_happened_at > self.last_sync_errored_at:
                self.last_sync_errored_at = action_happened_at
                update_fields.append('last_sync_errored_at')
            if self.last_content_sync_errored_at is None or action_happened_at > self.last_content_sync_errored_at:
                self.last_content_sync_errored_at = action_happened_at
                update_fields.append('last_content_sync_errored_at')
        if update_fields:
            return self.save(update_fields=update_fields)
        else:
            return self

    def update_learner_synced_at(self, action_happened_at, was_successful):
        """
        Given the last time a Learner record sync was attempted and status update the appropriate timestamps.
        """
        update_fields = []
        if self.last_sync_attempted_at is None or action_happened_at > self.last_sync_attempted_at:
            self.last_sync_attempted_at = action_happened_at
            update_fields.append('last_sync_attempted_at')
        if self.last_learner_sync_attempted_at is None or action_happened_at > self.last_learner_sync_attempted_at:
            self.last_learner_sync_attempted_at = action_happened_at
            update_fields.append('last_learner_sync_attempted_at')
        if not was_successful:
            if self.last_sync_errored_at is None or action_happened_at > self.last_sync_errored_at:
                self.last_sync_errored_at = action_happened_at
                update_fields.append('last_sync_errored_at')
            if self.last_learner_sync_errored_at is None or action_happened_at > self.last_learner_sync_errored_at:
                self.last_learner_sync_errored_at = action_happened_at
                update_fields.append('last_learner_sync_errored_at')
        if update_fields:
            return self.save(update_fields=update_fields)
        else:
            return self

    @property
    def is_valid(self):
        """
        Returns whether or not the configuration is valid and ready to be activated

        Args:
            obj: The instance of EnterpriseCustomerConfiguration
                being rendered with this admin form.
        """
        missing_items = {'missing': []}
        incorrect_items = {'incorrect': []}
        return missing_items, incorrect_items

    @property
    def channel_worker_user(self):
        """
        default worker username for channel
        """
        worker_username = self.channel_worker_username if self.channel_worker_username else 'enterprise_channel_worker'
        return User.objects.filter(username=worker_username).first()

    @property
    def customer_catalogs_to_transmit(self):
        """
        Return the list of EnterpriseCustomerCatalog objects.
        """
        catalogs_list = []
        if self.catalogs_to_transmit:
            catalogs_list = EnterpriseCustomerCatalog.objects.filter(
                uuid__in=convert_comma_separated_string_to_list(self.catalogs_to_transmit)
            )
        return catalogs_list

    @staticmethod
    def channel_code():
        """
        Returns an capitalized identifier for this channel class, unique among subclasses.
        """
        raise NotImplementedError('Implemented in concrete subclass.')

    @classmethod
    def get_class_by_channel_code(cls, channel_code):
        """
        Return the `EnterpriseCustomerPluginConfiguration` implementation for the particular channel_code, or None
        """
        for a_cls in cls.__subclasses__():
            if a_cls.channel_code().lower() == channel_code.lower():
                return a_cls
        return None

    def generate_default_display_name(self):
        """
        Returns a default display namem which can be overriden by a subclass.
        """
        return f'{self.channel_code()} {self.id}'

    def get_learner_data_exporter(self, user):
        """
        Returns the class that can serialize the learner course completion data to the integrated channel.
        """
        return LearnerExporter(user, self)

    def get_learner_data_transmitter(self):
        """
        Returns the class that can transmit the learner course completion data to the integrated channel.
        """
        return LearnerTransmitter(self)

    def get_content_metadata_exporter(self, user):
        """
        Returns a class that can retrieve and transform content metadata to the schema
        expected by the integrated channel.
        """
        return ContentMetadataExporter(user, self)

    def get_content_metadata_transmitter(self):
        """
        Returns a class that can transmit the content metadata to the integrated channel.
        """
        return ContentMetadataTransmitter(self)

    def transmit_learner_data(self, user):
        """
        Iterate over each learner data record and transmit it to the integrated channel.
        """
        exporter = self.get_learner_data_exporter(user)
        transmitter = self.get_learner_data_transmitter()
        transmitter.transmit(exporter)

    def transmit_single_learner_data(self, **kwargs):
        """
        Iterate over single learner data record and transmit it to the integrated channel.
        """
        exporter = self.get_learner_data_exporter(self.channel_worker_user)
        transmitter = self.get_learner_data_transmitter()
        transmitter.transmit(exporter, **kwargs)

    def transmit_content_metadata(self, user):
        """
        Transmit content metadata to integrated channel.
        """
        exporter = self.get_content_metadata_exporter(user)
        transmitter = self.get_content_metadata_transmitter()
        transmitter.transmit(*exporter.export())

    def transmit_single_subsection_learner_data(self, **kwargs):
        """
        Transmit a single subsection learner data record to the integrated channel.
        """
        exporter = self.get_learner_data_exporter(self.channel_worker_user)
        transmitter = self.get_learner_data_transmitter()
        transmitter.single_learner_assessment_grade_transmit(exporter, **kwargs)

    def transmit_subsection_learner_data(self, user):
        """
        Iterate over each assessment learner data record and transmit them to the integrated channel.
        """
        exporter = self.get_learner_data_exporter(user)
        transmitter = self.get_learner_data_transmitter()
        transmitter.assessment_level_transmit(exporter)

    def cleanup_duplicate_assignment_records(self, user):
        """
        Remove duplicated assessments transmitted through the integrated channel.
        """
        exporter = self.get_learner_data_exporter(user)
        transmitter = self.get_learner_data_transmitter()
        transmitter.deduplicate_assignment_records_transmit(exporter)

    def update_content_transmission_catalog(self, user):
        """
        Update transmission audits to contain the content's associated catalog uuid.
        """
        exporter = self.get_content_metadata_exporter(user)
        exporter.update_content_transmissions_catalog_uuids()


class GenericEnterpriseCustomerPluginConfiguration(EnterpriseCustomerPluginConfiguration):
    """
    A generic implementation of EnterpriseCustomerPluginConfiguration which can be instantiated

    .. no_pii:
    """

    def __str__(self):
        """
        Return human-readable string representation.
        """
        return "<GenericEnterpriseCustomerPluginConfiguration for Enterprise {enterprise_name}>".format(
            enterprise_name=self.enterprise_customer.name
        )

    @staticmethod
    def channel_code():
        """
        Returns an capitalized identifier for this channel class, unique among subclasses.
        """
        return 'GENERIC'


class ApiResponseRecord(TimeStampedModel):
    """
    Api response data for learner and content metadata transmissions

    .. no_pii:
    """
    status_code = models.PositiveIntegerField(
        help_text='The most recent remote API call response HTTP status code',
        blank=True,
        null=True
    )
    body = models.TextField(
        help_text='The most recent remote API call response body',
        blank=True,
        null=True
    )


class LearnerDataTransmissionAudit(TimeStampedModel):
    """
    The payload we send to an integrated channel  at a given point in time for an enterprise course enrollment.

    .. pii: The user_email model field contains PII. Declaring "retained" because I don't know if it's retired.
    .. pii_types: email_address
    .. pii_retirement: retained
    """

    # TODO: index customer uuid + plugin coinfig id together, with enrollment id?
    enterprise_customer_uuid = models.UUIDField(blank=True, null=True)
    user_email = models.CharField(max_length=255, blank=True, null=True)
    plugin_configuration_id = models.IntegerField(blank=True, null=True)
    enterprise_course_enrollment_id = models.IntegerField(blank=True, null=True, db_index=True)
    course_id = models.CharField(max_length=255, blank=False, null=False)
    content_title = models.CharField(max_length=255, default=None, null=True, blank=True)
    course_completed = models.BooleanField(default=True)
    progress_status = models.CharField(max_length=255, blank=True)
    completed_timestamp = models.DateTimeField(blank=True, null=True)
    instructor_name = models.CharField(max_length=255, blank=True)
    grade = models.FloatField(blank=True, null=True)
    total_hours = models.FloatField(null=True, blank=True)
    subsection_id = models.CharField(max_length=255, blank=True, null=True, db_index=True)
    subsection_name = models.CharField(max_length=255, blank=False, null=True)
    status = models.CharField(max_length=100, blank=True, null=True)
    error_message = models.TextField(blank=True, null=True)
    is_transmitted = models.BooleanField(default=False)
    friendly_status_message = models.CharField(
        help_text='A user-friendly API response status message.',
        max_length=255,
        default=None,
        null=True,
        blank=True
    )
    api_record = models.OneToOneField(
        ApiResponseRecord,
        blank=True,
        null=True,
        on_delete=models.CASCADE,
        help_text=_('Data pertaining to the transmissions API request response.')
    )

    class Meta:
        abstract = True
        app_label = 'integrated_channel'

    def __str__(self):
        """
        Return a human-readable string representation of the object.
        """
        return (
            f'<LearnerDataTransmissionAudit {self.id}'
            f' for enterprise enrollment {self.enterprise_course_enrollment_id}, '
            f', course_id: {self.course_id}>'
            f', grade: {self.grade}'
            f', completed_timestamp: {self.completed_timestamp}'
            f', enterprise_customer_uuid: {self.enterprise_customer_uuid}'
            f', course_completed: {self.course_completed}'
        )

    def __repr__(self):
        """
        Return uniquely identifying string representation.
        """
        return self.__str__()

    @property
    def provider_id(self):
        """
        Fetch ``provider_id`` from global configuration settings
        """
        return None

    @classmethod
    def audit_type(cls):
        """
        The base learner data transmission audit type - defaults to `completion`
        """
        return "completion"

    @classmethod
    def get_completion_class_by_channel_code(cls, channel_code):
        """
        Return the `LearnerDataTransmissionAudit` implementation related to completion reporting for
        the particular channel_code, or None
        """
        app_label = channel_code_to_app_label(channel_code)
        for a_cls in cls.__subclasses__():
            if a_cls._meta.app_label == app_label and a_cls.audit_type() == "completion":
                return a_cls
        return None

    def serialize(self, *args, **kwargs):
        """
        Return a JSON-serialized representation.

        Sort the keys so the result is consistent and testable.

        # TODO: When we refactor to use a serialization flow consistent with how course metadata
        # is serialized, remove the serialization here and make the learner data exporter handle the work.
        """
        return json.dumps(self._payload_data(), sort_keys=True)

    def _payload_data(self):
        """
        Convert the audit record's fields into SAP SuccessFactors key/value pairs.
        """
        return {
            'courseID': self.course_id,
            'courseCompleted': 'true' if self.course_completed else 'false',
            'completedTimestamp': self.completed_timestamp,
            'grade': self.grade,
        }


class GenericLearnerDataTransmissionAudit(LearnerDataTransmissionAudit):
    """
    A generic implementation of LearnerDataTransmissionAudit which can be instantiated
    """
    class Meta:
        app_label = 'integrated_channel'

    def __str__(self):
        """
        Return a human-readable string representation of the object.
        """
        return (
            '<GenericLearnerDataTransmissionAudit {transmission_id} for enterprise enrollment {enrollment}, '
            'and course {course_id}>'.format(
                transmission_id=self.id,
                enrollment=self.enterprise_course_enrollment_id,
                course_id=self.course_id
            )
        )

    def __repr__(self):
        """
        Return uniquely identifying string representation.
        """
        return self.__str__()


class ContentMetadataItemTransmission(TimeStampedModel):
    """
    A content metadata item that has been transmitted to an integrated channel.

    This model can be queried to find the content metadata items that have been
    transmitted to an integrated channel. It is used to synchronize the content
    metadata items available in an enterprise's catalog with the integrated channel.

    .. no_pii:
    """
    class Meta:
        indexes = [
            models.Index(
                fields=['enterprise_customer', 'integrated_channel_code', 'plugin_configuration_id', 'content_id'],
                name="customer_code_plugin_id_idx"
            ),
        ]
        unique_together = (('integrated_channel_code', 'plugin_configuration_id', 'content_id'),)

    enterprise_customer = models.ForeignKey(EnterpriseCustomer, on_delete=models.CASCADE)
    integrated_channel_code = models.CharField(max_length=30)
    plugin_configuration_id = models.PositiveIntegerField(blank=True, null=True)
    content_id = models.CharField(max_length=255)
    content_title = models.CharField(max_length=255, default=None, null=True, blank=True)
    channel_metadata = JSONField()
    content_last_changed = models.DateTimeField(
        help_text='Date of the last time the enterprise catalog associated with this metadata item was updated',
        blank=True,
        null=True
    )
    enterprise_customer_catalog_uuid = models.UUIDField(
        help_text='The enterprise catalog that this metadata item was derived from',
        blank=True,
        null=True,
    )
    remote_deleted_at = models.DateTimeField(
        help_text='Date when the content transmission was deleted in the remote API',
        blank=True,
        null=True
    )
    remote_created_at = models.DateTimeField(
        help_text='Date when the content transmission was created in the remote API',
        blank=True,
        null=True
    )
    remote_errored_at = models.DateTimeField(
        help_text='Date when the content transmission was failed in the remote API.',
        blank=True,
        null=True
    )
    remote_updated_at = models.DateTimeField(
        help_text='Date when the content transmission was last updated in the remote API',
        blank=True,
        null=True
    )
    api_response_status_code = models.PositiveIntegerField(
        help_text='The most recent remote API call response HTTP status code',
        blank=True,
        null=True
    )
    friendly_status_message = models.CharField(
        help_text='A user-friendly API response status message.',
        max_length=255,
        default=None,
        null=True,
        blank=True
    )
    marked_for = models.CharField(
        help_text='Flag marking a record as needing a form of transmitting',
        max_length=32,
        blank=True,
        null=True
    )
    api_record = models.OneToOneField(
        ApiResponseRecord,
        blank=True,
        null=True,
        on_delete=models.CASCADE,
        help_text=_('Data pertaining to the transmissions API request response.')
    )

    @classmethod
    def deleted_transmissions(cls, enterprise_customer, plugin_configuration_id, integrated_channel_code, content_id):
        """
        Return any pre-existing records for this customer/plugin/content which was previously deleted
        """
        query = Q(
            enterprise_customer=enterprise_customer,
            plugin_configuration_id=plugin_configuration_id,
            content_id=content_id,
            integrated_channel_code=integrated_channel_code,
            remote_deleted_at__isnull=False,
        )
        query.add(
            Q(remote_errored_at__lt=LAST_24_HRS)
            | Q(remote_errored_at__isnull=True)
            | Q(remote_errored_at__lt=enterprise_customer.modified), Q.AND)
        return ContentMetadataItemTransmission.objects.filter(query)

    @classmethod
    def incomplete_create_transmissions(
        cls,
        enterprise_customer,
        plugin_configuration_id,
        integrated_channel_code,
        content_id
    ):
        """
        Return any pre-existing records for this customer/plugin/content which was created but never sent or failed
        """
        in_db_but_unsent_query = Q(
            enterprise_customer=enterprise_customer,
            plugin_configuration_id=plugin_configuration_id,
            content_id=content_id,
            integrated_channel_code=integrated_channel_code,
            remote_created_at__isnull=True,
            remote_updated_at__isnull=True,
            remote_deleted_at__isnull=True,
            remote_errored_at__isnull=True,
        )
        in_db_but_failed_to_send_query = Q(
            enterprise_customer=enterprise_customer,
            plugin_configuration_id=plugin_configuration_id,
            content_id=content_id,
            integrated_channel_code=integrated_channel_code,
            remote_created_at__isnull=False,
            remote_updated_at__isnull=True,
            remote_deleted_at__isnull=True,
            api_response_status_code__gte=400,
        )
        in_db_but_failed_to_send_query.add(
            Q(remote_errored_at__lt=LAST_24_HRS) | Q(remote_errored_at__isnull=True) |
            Q(remote_errored_at__lt=enterprise_customer.modified), Q.AND)
        in_db_but_unsent_query.add(in_db_but_failed_to_send_query, Q.OR)
        return ContentMetadataItemTransmission.objects.filter(in_db_but_unsent_query)

    @classmethod
    def incomplete_update_transmissions(
        cls,
        enterprise_customer,
        plugin_configuration_id,
        integrated_channel_code,
        content_id
    ):
        """
        Return any pre-existing records for this customer/plugin/content which was updated but never sent or failed
        """
        in_db_but_failed_to_send_query = Q(
            enterprise_customer=enterprise_customer,
            plugin_configuration_id=plugin_configuration_id,
            content_id=content_id,
            integrated_channel_code=integrated_channel_code,
            remote_created_at__isnull=False,
            remote_updated_at__isnull=False,
            remote_deleted_at__isnull=True,
            api_response_status_code__gte=400,
        )
        in_db_but_failed_to_send_query.add(
            Q(remote_errored_at__lt=LAST_24_HRS) | Q(remote_errored_at__isnull=True) |
            Q(remote_errored_at__lt=enterprise_customer.modified), Q.AND)
        return ContentMetadataItemTransmission.objects.filter(in_db_but_failed_to_send_query)

    @classmethod
    def incomplete_delete_transmissions(
        cls,
        enterprise_customer,
        plugin_configuration_id,
        integrated_channel_code,
        content_id
    ):
        """
        Return any pre-existing records for this customer/plugin/content which was deleted but never sent or failed
        """
        in_db_but_failed_to_send_query = Q(
            enterprise_customer=enterprise_customer,
            plugin_configuration_id=plugin_configuration_id,
            content_id=content_id,
            integrated_channel_code=integrated_channel_code,
            remote_created_at__isnull=False,
            remote_deleted_at__isnull=False,
            api_response_status_code__gte=400,
        )
        in_db_but_failed_to_send_query.add(
            Q(remote_errored_at__lt=LAST_24_HRS)
            | Q(remote_errored_at__isnull=True)
            | Q(remote_errored_at__lt=enterprise_customer.modified), Q.AND)
        return ContentMetadataItemTransmission.objects.filter(in_db_but_failed_to_send_query)

    def _mark_transmission(self, mark_for):
        """
        Helper method to tag a transmission for any operation
        """
        self.marked_for = mark_for
        self.save()

    def mark_for_create(self):
        """
        Mark a transmission for creation
        """
        self._mark_transmission(TRANSMISSION_MARK_CREATE)

    def mark_for_update(self):
        """
        Mark a transmission for update
        """
        self._mark_transmission(TRANSMISSION_MARK_UPDATE)

    def mark_for_delete(self):
        """
        Mark a transmission for delete
        """
        self._mark_transmission(TRANSMISSION_MARK_DELETE)

    def remove_marked_for(self):
        """
        Remove and mark on a transmission
        """
        self._mark_transmission(None)

    def prepare_to_recreate(self, content_last_changed, enterprise_customer_catalog_uuid):
        """
        Prepare a deleted or unsent record to be re-created in the remote API by resetting dates and audit fields
        """
        # maintaining status code on the transmission record to aid with querying
        self.api_response_status_code = None
        if self.api_record:
            self.api_record.body = None
            self.api_record.status_code = None
        self.remote_deleted_at = None
        self.remote_created_at = None
        self.remote_updated_at = None
        self.channel_metadata = None
        self.content_last_changed = content_last_changed
        self.enterprise_customer_catalog_uuid = enterprise_customer_catalog_uuid
        self.marked_for = TRANSMISSION_MARK_CREATE
        self.save()
        return self

    def __str__(self):
        """
        Return a human-readable string representation of the object.
        """
        return (
            '<Content item {content_id} for Customer {customer} with Channel {channel}>'.format(
                content_id=self.content_id,
                customer=self.enterprise_customer,
                channel=self.integrated_channel_code
            )
        )

    def __repr__(self):
        """
        Return uniquely identifying string representation.
        """
        return self.__str__()


class OrphanedContentTransmissions(TimeStampedModel):
    """
    A model to track content metadata transmissions that were successfully sent to the integrated channel but then
    subsequently were orphaned by a removal of their associated catalog from the customer.

    .. no_pii:
    """
    class Meta:
        indexes = [
            models.Index(
                fields=['integrated_channel_code', 'plugin_configuration_id', 'resolved'],
                name="channel_plugin_resolved_idx"
            ),
        ]

    integrated_channel_code = models.CharField(max_length=30)
    plugin_configuration_id = models.PositiveIntegerField(blank=False, null=False)
    content_id = models.CharField(max_length=255, blank=False, null=False)
    transmission = models.ForeignKey(
        ContentMetadataItemTransmission,
        related_name='orphaned_record',
        on_delete=models.CASCADE,
    )
    resolved = models.BooleanField(default=False)


class IntegratedChannelAPIRequestLogs(TimeStampedModel):
    """
    A model to track basic information about every API call we make from the integrated channels.

    .. no_pii:
    """

    enterprise_customer = models.ForeignKey(
        EnterpriseCustomer, on_delete=models.CASCADE
    )
    enterprise_customer_configuration_id = models.IntegerField(
        blank=False,
        null=False,
        help_text="ID from the EnterpriseCustomerConfiguration model",
    )
    endpoint = models.URLField(
        blank=False,
        max_length=255,
        null=False,
    )
    payload = models.TextField(blank=False, null=False)
    time_taken = models.FloatField(blank=False, null=False)
    status_code = models.PositiveIntegerField(
        help_text="API call response HTTP status code", blank=True, null=True
    )
    response_body = models.TextField(
        help_text="API call response body", blank=True, null=True
    )
    channel_name = models.TextField(
        help_text="Name of the integrated channel associated with this API call log record.",
        blank=True
    )

    class Meta:
        app_label = "integrated_channel"
        verbose_name_plural = "Integrated channels API request logs"

    def __str__(self):
        """
        Return a human-readable string representation of the object.
        """
        return (
            f"<IntegratedChannelAPIRequestLog {self.id}"
            f" for enterprise customer {self.enterprise_customer} "
            f", enterprise_customer_configuration_id: {self.enterprise_customer_configuration_id}>"
            f", endpoint: {self.endpoint}"
            f", time_taken: {self.time_taken}"
            f", response_body: {self.response_body}"
            f", status_code: {self.status_code}"
        )

    def __repr__(self):
        """
        Return uniquely identifying string representation.
        """
        return self.__str__()

    @classmethod
    def store_api_call(
        cls,
        enterprise_customer,
        enterprise_customer_configuration_id,
        endpoint,
        payload,
        time_taken,
        status_code,
        response_body,
        channel_name
    ):
        """
        Creates new record in IntegratedChannelAPIRequestLogs table.
        """
        try:
            record = cls(
                enterprise_customer=enterprise_customer,
                enterprise_customer_configuration_id=enterprise_customer_configuration_id,
                endpoint=endpoint,
                payload=payload,
                time_taken=time_taken,
                status_code=status_code,
                response_body=response_body,
                channel_name=channel_name
            )
            record.save()
        except Exception as e:  # pylint: disable=broad-except
            LOGGER.error(
                f"store_api_call raised error while storing API call: {e}"
                f"enterprise_customer={enterprise_customer}"
                f"enterprise_customer_configuration_id={enterprise_customer_configuration_id},"
                f"endpoint={endpoint}"
                f"payload={payload}"
                f"time_taken={time_taken}"
                f"status_code={status_code}"
                f"response_body={response_body}"
                f"channel_name={channel_name}"
            )



# Generated by Django 4.2.16 on 2024-10-23 22:38

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('enterprise', '0224_alter_enterprisegroup_applies_to_all_contexts_and_more'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='enterprisegroup',
            name='applies_to_all_contexts',
        ),
        migrations.RemoveField(
            model_name='historicalenterprisegroup',
            name='applies_to_all_contexts',
        ),
        migrations.AlterField(
            model_name='enterprisegroup',
            name='name',
            field=models.CharField(help_text='Specifies enterprise group name.', max_length=250),
        ),
        migrations.AlterField(
            model_name='historicalenterprisegroup',
            name='name',
            field=models.CharField(help_text='Specifies enterprise group name.', max_length=250),
        ),
    ]



from django.conf import settings
from django.db import migrations, models, connection
import django.db.models.deletion

mysql_sql = """
                    SET FOREIGN_KEY_CHECKS = 0;

                    ALTER TABLE enterprise_enterprisecustomeruser
                    ADD CONSTRAINT fk_enterprisecustomeruser_user_fk
                    FOREIGN KEY (user_fk) REFERENCES auth_user(id)
                    ON DELETE CASCADE,
                    ALGORITHM=INPLACE,
                    LOCK=NONE;

                    ALTER TABLE enterprise_historicalenterprisecustomeruser
                    ADD CONSTRAINT fk_historical_enterprisecustomeruser_user_fk
                    FOREIGN KEY (user_fk) REFERENCES auth_user(id)
                    ON DELETE CASCADE,
                    ALGORITHM=INPLACE,
                    LOCK=NONE;

                    SET FOREIGN_KEY_CHECKS = 1;
            """
psql_sql = """

           ALTER TABLE enterprise_enterprisecustomeruser
               ADD CONSTRAINT fk_enterprisecustomeruser_user_fk
                   FOREIGN KEY (user_fk) REFERENCES auth_user (id)
                       ON DELETE CASCADE;

           ALTER TABLE enterprise_historicalenterprisecustomeruser
               ADD CONSTRAINT fk_historical_enterprisecustomeruser_user_fk
                   FOREIGN KEY (user_fk) REFERENCES auth_user (id)
                       ON DELETE CASCADE;

           """
mysql_reverse_sql = """
                    ALTER TABLE enterprise_historicalenterprisecustomeruser
                        DROP FOREIGN KEY fk_historical_enterprisecustomeruser_user_fk;

                    ALTER TABLE enterprise_enterprisecustomeruser
                        DROP FOREIGN KEY fk_enterprisecustomeruser_user_fk; \
                    """
psql_reverse_sql = """
                   ALTER TABLE enterprise_historicalenterprisecustomeruser
                       DROP CONSTRAINT fk_historical_enterprisecustomeruser_user_fk;

                   ALTER TABLE enterprise_enterprisecustomeruser
                       DROP CONSTRAINT fk_enterprisecustomeruser_user_fk; \
                   """


class Migration(migrations.Migration):
    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('enterprise', '0230_alter_enterprisecustomeruser_user_fk_and_more'),
    ]

    db_engine = connection.settings_dict['ENGINE']
    if 'sqlite3' in db_engine:
        operations = []
    else:
        operations = [
            migrations.RunSQL(
                sql=psql_sql if 'postgres' in db_engine else mysql_sql,
                reverse_sql=psql_reverse_sql if 'postgres' in db_engine else mysql_reverse_sql,
                state_operations=[
                    migrations.AlterField(
                        model_name='enterprisecustomeruser',
                        name='user_fk',
                        field=models.ForeignKey(
                            null=True,
                            on_delete=django.db.models.deletion.CASCADE,
                            related_name='enterprise_customer_users',
                            to=settings.AUTH_USER_MODEL,
                        ),
                    ),
                    migrations.AlterField(
                        model_name='historicalenterprisecustomeruser',
                        name='user_fk',
                        field=models.ForeignKey(
                            null=True,
                            on_delete=django.db.models.deletion.CASCADE,
                            related_name='+',  # Historical tables often use '+' for no backward relation
                            to=settings.AUTH_USER_MODEL,
                        ),
                    ),
                ],
            ),
        ]



# Generated by Django 3.2.23 on 2024-04-23 10:57

from django.db import migrations
from integrated_channels.blackboard.utils import populate_decrypted_fields_blackboard


class Migration(migrations.Migration):

    dependencies = [
        ('blackboard', '0020_auto_20240422_1709'),
    ]

    operations = [
        migrations.RunPython(populate_decrypted_fields_blackboard, reverse_code=migrations.RunPython.noop),
    ]



# Generated by Django 3.2.20 on 2023-08-29 17:41

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('enterprise', '0181_enterprisecustomerssoconfiguration_historicalenterprisecustomerssoconfiguration'),
    ]

    operations = [
        migrations.AddField(
            model_name='enterprisecustomer',
            name='enable_generation_of_api_credentials',
            field=models.BooleanField(default=False, verbose_name='Allow generation of API credentials'),
        ),
        migrations.AddField(
            model_name='historicalenterprisecustomer',
            name='enable_generation_of_api_credentials',
            field=models.BooleanField(default=False, verbose_name='Allow generation of API credentials'),
        ),
    ]



from django.db import connection, migrations


class Migration(migrations.Migration):
    dependencies = [
        ('enterprise', '0178_auto_20230808_0923'),
    ]

    # At one point we ran into migration deployment issues that caused us to manually change and alter tables on stage
    # and prod. One unintended consequence of this event was that the historical licensed enrollment table was dropped
    # and never restored on __stage__ only. This migration restores the historical licensed enrollment table IFF it
    # does not exist as to bypass the error that would occur if we tried to create a table that already exists.
    # Relevant RCA: https://2u-internal.atlassian.net/wiki/spaces/ENG/pages/345374826/RCA+Enterprise+partially-applied+licensed+course+enrollment+migration
    db_engine = connection.settings_dict['ENGINE']
    if 'mysql' in db_engine:
        operations = [
            migrations.RunSQL(
                """
                CREATE TABLE IF NOT EXISTS `enterprise_historicallicensedenterprisecourseenrollment` (
                `id` int(11) NOT NULL,
                `created` datetime(6) NOT NULL,
                `modified` datetime(6) NOT NULL,
                `license_uuid` char(32) NOT NULL,
                `history_id` int(11) NOT NULL AUTO_INCREMENT,
                `history_date` datetime(6) NOT NULL,
                `history_change_reason` varchar(100) DEFAULT NULL,
                `history_type` varchar(1) NOT NULL,
                `enterprise_course_enrollment_id` int(11) DEFAULT NULL,
                `history_user_id` int(11) DEFAULT NULL,
                `is_revoked` tinyint(1) NOT NULL,
                PRIMARY KEY (`history_id`),
                KEY `enterprise_historica_history_user_id_1db87766_fk_auth_user` (`history_user_id`),
                KEY `enterprise_historicallicens_id_ff4cfd4f` (`id`),
                KEY `enterprise_historicallicens_enterprise_course_enrollmen_1b0d3427` (`enterprise_course_enrollment_id`),
                CONSTRAINT `enterprise_historica_history_user_id_1db87766_fk_auth_user` FOREIGN KEY (`history_user_id`) REFERENCES `auth_user` (`id`)
                );
                """,
                reverse_sql=migrations.RunSQL.noop,
            ),
        ]



# Generated by Django 3.2.11 on 2022-01-31 15:39

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('cornerstone', '0013_auto_20220126_1837'),
    ]

    operations = [
        migrations.AddField(
            model_name='cornerstoneenterprisecustomerconfiguration',
            name='display_name',
            field=models.CharField(blank=True, default='', help_text='A configuration nickname.', max_length=30),
        ),
        migrations.AddField(
            model_name='historicalcornerstoneenterprisecustomerconfiguration',
            name='display_name',
            field=models.CharField(blank=True, default='', help_text='A configuration nickname.', max_length=30),
        ),
    ]



# Generated by Django 4.2.15 on 2024-09-02 09:43

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('canvas', '0038_alter_canvasenterprisecustomerconfiguration_client_id_and_more'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='canvasenterprisecustomerconfiguration',
            name='client_id',
        ),
        migrations.RemoveField(
            model_name='canvasenterprisecustomerconfiguration',
            name='client_secret',
        ),
    ]



# Generated by Django 3.2.23 on 2024-09-09 15:56

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('sap_success_factors', '0022_auto_20240906_1349'),
    ]

    operations = [
        migrations.AlterField(
            model_name='sapsuccessfactorsenterprisecustomerconfiguration',
            name='key',
            field=models.CharField(blank=True, default='', help_text='OAuth client identifier.', max_length=255, null=True, verbose_name='Client ID'),
        ),
        migrations.AlterField(
            model_name='sapsuccessfactorsenterprisecustomerconfiguration',
            name='secret',
            field=models.CharField(blank=True, default='', help_text='OAuth client secret.', max_length=255, null=True, verbose_name='Client Secret'),
        ),
    ]



# -*- coding: utf-8 -*-
"""
Client for connecting to Degreed2.
"""

import json
import logging
import time
from http import HTTPStatus

import requests
from edx_django_utils.cache import TieredCache, get_cache_key
from six.moves.urllib.parse import urljoin

from django.apps import apps
from django.conf import settings
from django.http.request import QueryDict

from enterprise.api_client.enterprise_catalog import EnterpriseCatalogApiClient
from enterprise.models import EnterpriseCustomerUser
from integrated_channels.exceptions import ClientError
from integrated_channels.integrated_channel.client import IntegratedChannelApiClient
from integrated_channels.utils import generate_formatted_log, refresh_session_if_expired, stringify_and_store_api_record

LOGGER = logging.getLogger(__name__)


class Degreed2APIClient(IntegratedChannelApiClient):
    """
    Client for connecting to Degreed2.

    Specifically, this class supports obtaining access tokens and posting to the courses and
    completion status endpoints.
    """

    CONTENT_WRITE_SCOPE = "content:write"
    ALL_DESIRED_SCOPES = "content:read,content:write,completions:write,completions:read"
    SESSION_TIMEOUT = getattr(settings, "ENTERPRISE_DEGREED2_SESSION_TIMEOUT", 60)
    MAX_RETRIES = getattr(settings, "ENTERPRISE_DEGREED2_MAX_RETRIES", 4)
    BACKOFF_FACTOR = getattr(settings, "ENTERPRISE_DEGREED2_BACKOFF_FACTOR", 2)

    def __init__(self, enterprise_configuration):
        """
        Instantiate a new client.

        Args:
            enterprise_configuration (Degreed2EnterpriseCustomerConfiguration): An enterprise customers's
            configuration model for connecting with Degreed
        """
        super().__init__(enterprise_configuration)
        self.session = None
        self.expires_at = None
        app_config = apps.get_app_config('degreed2')
        self.oauth_api_path = app_config.oauth_api_path
        self.courses_api_path = app_config.courses_api_path
        self.completions_api_path = app_config.completions_api_path
        self.skill_api_path = app_config.skill_api_path
        # to log without having to pass channel_name, ent_customer_uuid each time
        self.make_log_msg = lambda course_key, message, lms_user_id=None: generate_formatted_log(
            self.enterprise_configuration.channel_code(),
            self.enterprise_configuration.enterprise_customer.uuid,
            lms_user_id,
            course_key,
            message,
        )
        self.enterprise_catalog_api_client = EnterpriseCatalogApiClient()
        self.IntegratedChannelAPIRequestLogs = apps.get_model(
            "integrated_channel", "IntegratedChannelAPIRequestLogs"
        )

    def get_oauth_url(self):
        config = self.enterprise_configuration
        base_url = config.degreed_token_fetch_base_url or config.degreed_base_url
        return urljoin(base_url, self.oauth_api_path)

    def get_courses_url(self):
        return urljoin(self.enterprise_configuration.degreed_base_url, self.courses_api_path)

    def get_completions_url(self):
        return urljoin(self.enterprise_configuration.degreed_base_url, self.completions_api_path)

    def get_course_skills_url(self, course_key):
        return urljoin(
            self.enterprise_configuration.degreed_base_url,
            self.skill_api_path.format(contentId=course_key)
        )

    def create_assessment_reporting(self, user_id, payload):
        """
        Not implemented yet.
        """
        LOGGER.error(
            generate_formatted_log(
                self.enterprise_configuration.channel_code(),
                self.enterprise_configuration.enterprise_customer.uuid,
                None,
                None,
                "Degreed2 integrated channel does not yet support assessment reporting."
            )
        )

    def cleanup_duplicate_assignment_records(self, courses):
        """
        Not implemented yet.
        """
        LOGGER.error(
            generate_formatted_log(
                self.enterprise_configuration.channel_code(),
                self.enterprise_configuration.enterprise_customer.uuid,
                None,
                None,
                "Degreed2 integrated channel does not yet support assignment deduplication."
            )
        )

    def create_course_completion(self, user_id, payload):
        """
        Completions status for a learner
        Ref: https://api.degreed.com/docs/#create-a-new-completion

        Arguments:
            - user_id: learner's id
            - payload: (dict) with keys:
                 {
                     degreed_user_email,
                     course_id,
                     completed_timestamp, (in the format "2018-08-01T00:00:00")
                 }
        Returns: status_code, response_text
        """
        json_payload = json.loads(payload)
        code, body = self._post(
            self.get_completions_url(),
            json_payload,
            self.ALL_DESIRED_SCOPES
        )
        if code == HTTPStatus.BAD_REQUEST.value:
            error_response = json.loads(body)
            for error in error_response['errors']:
                if 'detail' in error and 'Invalid user identifier' in error['detail']:
                    try:
                        enterprise_customer = self.enterprise_configuration.enterprise_customer
                        LOGGER.info(
                            generate_formatted_log(
                                self.enterprise_configuration.channel_code(),
                                self.enterprise_configuration.enterprise_customer.uuid,
                                None,
                                None,
                                f'User {user_id} was deleted on degreed side,'
                                f"so marking it as inactive and unlinking from enterprise"
                            )
                        )
                        # Unlink user from related Enterprise Customer
                        EnterpriseCustomerUser.objects.unlink_user(
                            enterprise_customer=enterprise_customer,
                            user_email=json_payload.get('data').get('attributes').get('user-id'),
                        )
                    except Exception as e:  # pylint: disable=broad-except
                        LOGGER.error(
                            generate_formatted_log(
                                self.enterprise_configuration.channel_code(),
                                self.enterprise_configuration.enterprise_customer.uuid,
                                None,
                                None,
                                f'Error occurred while unlinking a degreed2 learner: {user_id}. '
                                f'Payload: {json_payload}, Error: {e}'
                            )
                        )
        return code, body

    def delete_course_completion(self, user_id, payload):
        """
        Not implemented yet. deletion requires completion ID. We don't have a way to get the ID right now.
        So we may need to store id on our side to be able to do this.
        https://api.degreed.com/docs/#delete-a-specific-completion
        """
        LOGGER.error(
            generate_formatted_log(
                self.enterprise_configuration.channel_code(),
                self.enterprise_configuration.enterprise_customer.uuid,
                None,
                None,
                "Degreed2 integrated channel does not yet support deleting course completions."
            )
        )

    def fetch_degreed_course_id(self, external_id):
        """
        Fetch the 'id' of a course from cache first and if not found then send a request to Degreed2,
        given the external-id as a search param 'external-id' is the edX course key.
        """
        cache_key = get_cache_key(
            resource='degreed2_course_id',
            resource_id=external_id,
        )
        cached_course_id = TieredCache.get_cached_response(cache_key)
        if cached_course_id.is_found:
            LOGGER.info(self.make_log_msg(external_id, f'Found cached course id: {cached_course_id.value}'))
            return cached_course_id.value
        # QueryDict converts + to space
        params = QueryDict(f"filter[external_id]={external_id.replace('+', '%2B')}")
        course_search_url = f'{self.get_courses_url()}?{params.urlencode(safe="[]")}'
        LOGGER.info(self.make_log_msg(external_id, f'Attempting find course via url: {course_search_url}'))
        status_code, response_body = self._get(
            course_search_url,
            self.ALL_DESIRED_SCOPES
        )
        if status_code >= 400:
            raise ClientError(
                f'Degreed2: fetch request failed: attempted, external ID={external_id},'
                f'received status_code={status_code}',
                status_code=status_code
            )
        response_json = json.loads(response_body)
        if response_json['data']:
            # cache the course id with a 1 day expiration
            response_course_id = response_json['data'][0]['id']
            expires_in = 60 * 60 * 24  # 1 day
            TieredCache.set_all_tiers(cache_key, response_course_id, expires_in)
            return response_course_id
        raise ClientError(
            f'Degreed2: Attempted to find degreed course id but failed, external id was {external_id}'
            f', Response from Degreed was {response_body}')

    def assign_course_skills(self, course_id, serialized_data):  # pylint: disable=inconsistent-return-statements
        """
        Assign skills to a course.

        Args:
            course_id: Course key
            serialized_data: JSON-encoded object containing skills metadata.

        Raises:
            ClientError:
                If Degreed course id doesn't exist.
                If Degreed course skills API request fails.
        """

        degreed_course_id = self.fetch_degreed_course_id(course_id)
        if not degreed_course_id:
            raise ClientError(f'Degreed2: Cannot find course via external-id {course_id}')

        course_skills_url = self.get_course_skills_url(degreed_course_id)
        LOGGER.info(self.make_log_msg(course_id, f'Attempting to assign course skills {course_skills_url}'))
        try:
            status_code, response_body = self._patch(course_skills_url, serialized_data, self.CONTENT_WRITE_SCOPE)
            if status_code == 201:
                LOGGER.info(
                    self.make_log_msg(
                        course_id,
                        f'Succesfully assigned skills to course {course_id}')
                )
                return status_code, response_body
            elif status_code >= 400:
                raise ClientError(
                    f'Degreed2APIClient failed to assign skills to course {course_id}.'
                    f'Failed with status_code={status_code} and response={response_body}',
                )
        except requests.exceptions.RequestException as exc:
            raise ClientError(
                'Degreed2APIClient request to assign skills failed: {error} {message}'.format(
                    error=exc.__class__.__name__,
                    message=str(exc)
                )
            ) from exc

    def create_content_metadata(self, serialized_data):
        """
        Create content metadata using the Degreed course content API.

        Args:
            serialized_data: JSON-encoded object containing content metadata.

        Raises:
            ClientError: If Degreed API request fails.
        """
        channel_metadata_item = json.loads(serialized_data.decode('utf-8'))
        # only expect one course in this array as of now (chunk size is 1)
        a_course = channel_metadata_item['courses'][0]
        external_id = a_course.get('external-id')
        status_code, response_body = self._sync_content_metadata(a_course, 'post', self.get_courses_url())
        if status_code == 409:
            # course already exists, don't raise failure, but try to mark it as active on Degreed side
            # if succeeds, we'll treat this as a success
            LOGGER.warning(
                self.make_log_msg(
                    external_id,
                    f'Course with integration_id = {external_id} already exists, marking it as active'
                )
            )
            try:
                channel_metadata_item['courses'][0]['obsolete'] = False
                return self.update_content_metadata(json.dumps(channel_metadata_item).encode('utf-8'))
            except requests.exceptions.RequestException as exc:
                raise ClientError(
                    'Degreed2APIClient request failed while handling 409: {error} {message}'.format(
                        error=exc.__class__.__name__,
                        message=str(exc)
                    )
                ) from exc
        elif status_code >= 400:
            raise ClientError(
                f'Degreed2APIClient create_content_metadata failed with status {status_code}: {response_body}',
                status_code=status_code
            )
        self._fetch_and_assign_skills_to_course(external_id)

        return status_code, response_body

    def update_content_metadata(self, serialized_data):
        """
        Update content metadata using the Degreed course content API.

        Args:
            serialized_data: JSON-encoded object containing content metadata.

        Raises:
            ClientError: If Degreed API request fails.
        """
        channel_metadata_item = json.loads(serialized_data.decode('utf-8'))
        course_item = channel_metadata_item['courses'][0]
        external_id = course_item.get('external-id')

        course_id = self.fetch_degreed_course_id(external_id)
        if not course_id:
            raise ClientError(f'Degreed2: Cannot find course via external-id {external_id}')

        patch_url = f'{self.get_courses_url()}/{course_id}'
        LOGGER.info(self.make_log_msg(external_id, f'Attempting course update via {patch_url}'))
        patch_status_code, patch_response_body = self._sync_content_metadata(
            course_item,
            'patch',
            patch_url,
            course_id
        )

        self._fetch_and_assign_skills_to_course(external_id)

        return patch_status_code, patch_response_body

    def _fetch_and_assign_skills_to_course(self, external_id):
        """
        Fetches content metadata(skills) from enterprise catalog API
        and transmits them to Degreed2 against given external_id(course_id)

        Args:
            external_id: Course id that is assigned to a course on Degreed side
        """
        # We need to do 2 steps here:
        # 1. Fetch skills from enterprise-catalog

        metadata = self.enterprise_catalog_api_client.get_customer_content_metadata_content_identifier(
            enterprise_uuid=self.enterprise_configuration.enterprise_customer.uuid,
            content_id=external_id)
        LOGGER.info(
            generate_formatted_log(
                self.enterprise_configuration.channel_code(),
                self.enterprise_configuration.enterprise_customer.uuid,
                None,
                None,
                f"[Degreed2Client] metadata: {metadata}",
            )
        )

        # 2. Transmit to degreed
        skills = metadata.get("skill_names", [])
        if skills:
            try:
                self.assign_course_skills(external_id, skills)
            except ClientError as err:
                generate_formatted_log(
                    self.enterprise_configuration.channel_code(),
                    self.enterprise_configuration.enterprise_customer.uuid,
                    None,
                    None,
                    f"[Degreed2Client]: {err.message}",
                )

    def delete_content_metadata(self, serialized_data):
        """
        Delete content metadata using the Degreed course content API.

        Args:
            serialized_data: JSON-encoded object containing content metadata.

        Raises:
            ClientError: If Degreed API request fails.
        """
        # {{base_api_url}}/api/v2/content/courses?filter[external_id]=course-v1%3AedX%2B444555666%2B3T2021
        channel_metadata_item = json.loads(serialized_data.decode('utf-8'))
        course_item = channel_metadata_item['courses'][0]
        external_id = course_item.get('external-id')

        del_status_code = None
        del_response_body = ''
        try:
            course_id = self.fetch_degreed_course_id(external_id)
            if not course_id:
                raise ClientError(f'Degreed2: Cannot find course via external-id {external_id}')

            del_url = f'{self.get_courses_url()}/{course_id}'
            LOGGER.info(self.make_log_msg(external_id, f'Attempting course delete via {del_url}'))
            del_status_code, del_response_body = self._delete(
                del_url,
                None,
                self.ALL_DESIRED_SCOPES
            )
            if del_status_code >= 400:
                raise ClientError(
                    f'Degreed2: delete request failed: attempted, external ID={external_id},'
                    f'status_code={del_status_code}'
                    f'response_body={del_response_body}',
                    status_code=del_status_code
                )
        except requests.exceptions.RequestException as exc:
            raise ClientError(
                'Degreed2APIClient delete request failed: {error} {message}'.format(
                    error=exc.__class__.__name__,
                    message=str(exc)
                )
            ) from exc

        return del_status_code, del_response_body

    def _sync_content_metadata(self, course_attributes, http_method, override_url, degreed_course_id=None):
        """
        Synchronize content metadata using the Degreed course content API.
        Invokes the Create/Update operations of degreed api
        Used for create course, and update course use cases
        The json sent contains

        Args:
            course_attributes: JSON object containing content metadata converted into
              Degreed2 'attributes' in the payload.
            http_method: The HTTP method to use for the API request.
            override_url: uses this url to post to
            course_id: used to append at the end of the url as /ID if provided

        Raises:
            ClientError: If Degreed API request fails.
        """
        json_to_send = {
            "data": {
                "type": "content/courses",
                "attributes": course_attributes,
            }
        }

        # useful for update use case
        if degreed_course_id:
            json_to_send['data']['id'] = degreed_course_id

        try:
            status_code, response_body = getattr(self, '_' + http_method)(
                override_url,
                json_to_send,
                self.ALL_DESIRED_SCOPES
            )
        except requests.exceptions.RequestException as exc:
            raise ClientError(
                'Degreed2APIClient request failed: {error} {message}'.format(
                    error=exc.__class__.__name__,
                    message=str(exc)
                )
            ) from exc
        return status_code, response_body

    def _calculate_backoff(self, attempt_count):
        """
        Calculate the seconds to sleep based on attempt_count
        """
        return (self.BACKOFF_FACTOR * (2 ** (attempt_count - 1)))

    def _get(self, url, scope):
        """
        Make a GET request using the session object to a Degreed2 endpoint.

        Args:
            url (str): The url to send a GET request to.
            scope (str): Must be one of the scopes Degreed expects:
                        - `CONTENT_WRITE_SCOPE`
                        - `CONTENT_READ_SCOPE`
        """
        self._create_session(scope)
        attempts = 0
        while True:
            attempts = attempts + 1
            start_time = time.time()
            response = self.session.get(url)
            duration_seconds = time.time() - start_time
            self.IntegratedChannelAPIRequestLogs.store_api_call(
                enterprise_customer=self.enterprise_configuration.enterprise_customer,
                enterprise_customer_configuration_id=self.enterprise_configuration.id,
                endpoint=url,
                payload='',
                time_taken=duration_seconds,
                status_code=response.status_code,
                response_body=response.text,
                channel_name=self.enterprise_configuration.channel_code()
            )
            if attempts <= self.MAX_RETRIES and response.status_code == 429:
                sleep_seconds = self._calculate_backoff(attempts)
                LOGGER.warning(
                    generate_formatted_log(
                        self.enterprise_configuration.channel_code(),
                        self.enterprise_configuration.enterprise_customer.uuid,
                        None,
                        None,
                        f'[Degreed2Client]._get 429 detected from {url}, backing-off before retrying, '
                        f'sleeping {sleep_seconds} seconds...'
                    )
                )
                time.sleep(sleep_seconds)
            else:
                break
        return response.status_code, response.text

    def _post(self, url, data, scope):
        """
        Make a POST request using the session object to a Degreed2 endpoint.

        Args:
            url (str): The url to send a POST request to.
            data (dict): The json payload to POST.
            scope (str): Must be one of the scopes Degreed expects:
                        - `CONTENT_WRITE_SCOPE`
                        - `CONTENT_READ_SCOPE`
        """
        self._create_session(scope)
        attempts = 0
        while True:
            attempts = attempts + 1
            start_time = time.time()
            response = self.session.post(url, json=data)
            duration_seconds = time.time() - start_time
            stringify_and_store_api_record(
                enterprise_customer=self.enterprise_configuration.enterprise_customer,
                enterprise_customer_configuration_id=self.enterprise_configuration.id,
                endpoint=url,
                data=data,
                time_taken=duration_seconds,
                status_code=response.status_code,
                response_body=response.text,
                channel_name=self.enterprise_configuration.channel_code()
            )
            if attempts <= self.MAX_RETRIES and response.status_code == 429:
                sleep_seconds = self._calculate_backoff(attempts)
                LOGGER.warning(
                    generate_formatted_log(
                        self.enterprise_configuration.channel_code(),
                        self.enterprise_configuration.enterprise_customer.uuid,
                        None,
                        None,
                        f'[Degreed2Client]._post 429 detected from {url}, backing-off before retrying, '
                        f'sleeping {sleep_seconds} seconds...'
                    )
                )
                time.sleep(sleep_seconds)
            else:
                break
        return response.status_code, response.text

    def _patch(self, url, data, scope):
        """
        Make a PATCH request using the session object to a Degreed2 endpoint.

        Args:
            url (str): The url to send a POST request to.
            data (str): The json payload to POST.
            scope (str): Must be one of the scopes Degreed expects:
                        - `CONTENT_WRITE_SCOPE`
                        - `CONTENT_READ_SCOPE`
        """
        self._create_session(scope)
        attempts = 0
        while True:
            attempts = attempts + 1
            start_time = time.time()
            response = self.session.patch(url, json=data)
            duration_seconds = time.time() - start_time
            stringify_and_store_api_record(
                enterprise_customer=self.enterprise_configuration.enterprise_customer,
                enterprise_customer_configuration_id=self.enterprise_configuration.id,
                endpoint=url,
                data=data,
                time_taken=duration_seconds,
                status_code=response.status_code,
                response_body=response.text,
                channel_name=self.enterprise_configuration.channel_code()
            )
            if attempts <= self.MAX_RETRIES and response.status_code == 429:
                sleep_seconds = self._calculate_backoff(attempts)
                LOGGER.warning(
                    generate_formatted_log(
                        self.enterprise_configuration.channel_code(),
                        self.enterprise_configuration.enterprise_customer.uuid,
                        None,
                        None,
                        f'429 detected from {url}, backing-off before retrying, '
                        f'sleeping {sleep_seconds} seconds...'
                    )
                )
                time.sleep(sleep_seconds)
            else:
                break
        return response.status_code, response.text

    def _delete(self, url, data, scope):
        """
        Make a DELETE request using the session object to a Degreed endpoint.

        Args:
            url (str): The url to send a DELETE request to.
            data (str): The json payload to DELETE. None means nothing is sent in payload.
            scope (str): Must be one of the scopes Degreed expects:
                        - `CONTENT_PROVIDER_SCOPE`
                        - `COMPLETION_PROVIDER_SCOPE`
        """
        self._create_session(scope)
        attempts = 0
        while True:
            attempts = attempts + 1
            start_time = time.time()
            response = self.session.delete(url, json=data) if data else self.session.delete(url)
            duration_seconds = time.time() - start_time
            stringify_and_store_api_record(
                enterprise_customer=self.enterprise_configuration.enterprise_customer,
                enterprise_customer_configuration_id=self.enterprise_configuration.id,
                endpoint=url,
                data=data if data else '',
                time_taken=duration_seconds,
                status_code=response.status_code,
                response_body=response.text,
                channel_name=self.enterprise_configuration.channel_code()
            )
            if attempts <= self.MAX_RETRIES and response.status_code == 429:
                sleep_seconds = self._calculate_backoff(attempts)
                LOGGER.warning(
                    generate_formatted_log(
                        self.enterprise_configuration.channel_code(),
                        self.enterprise_configuration.enterprise_customer.uuid,
                        None,
                        None,
                        f'429 detected from {url}, backing-off before retrying, '
                        f'sleeping {sleep_seconds} seconds...'
                    )
                )
                time.sleep(sleep_seconds)
            else:
                break
        return response.status_code, response.text

    def _create_session(self, scope):
        """
        Instantiate a new session object for use in connecting with Degreed
        """
        self.session, self.expires_at = refresh_session_if_expired(
            lambda: self._get_oauth_access_token(scope),
            self.session,
            self.expires_at,
        )

    def _get_oauth_access_token(self, scope):
        """ Retrieves OAuth 2.0 access token using the client credentials grant.
        Prefers using the degreed_token_fetch_base_url over the degreed_base_url, if present, to fetch the access token.

        Args:
            scope (str): Must be one or comma separated list of the scopes Degreed expects
        Returns:
            tuple: Tuple containing access token string and expiration datetime.
        Raises:
            HTTPError: If we received a failure response code from Degreed.
            ClientError: If an unexpected response format was received that we could not parse.
        """
        config = self.enterprise_configuration
        url = self.get_oauth_url()
        data = {
            'grant_type': 'client_credentials',
            'scope': scope,
            'client_id': config.decrypted_client_id,
            'client_secret': config.decrypted_client_secret,
        }
        start_time = time.time()
        response = requests.post(
            url,
            data=data,
            headers={'Content-Type': 'application/x-www-form-urlencoded'}
        )
        duration_seconds = time.time() - start_time
        stringify_and_store_api_record(
            enterprise_customer=self.enterprise_configuration.enterprise_customer,
            enterprise_customer_configuration_id=self.enterprise_configuration.id,
            endpoint=url,
            data=data,
            time_taken=duration_seconds,
            status_code=response.status_code,
            response_body=response.text,
            channel_name=self.enterprise_configuration.channel_code()
        )

        try:
            data = response.json()
            return data['access_token'], data['expires_in']
        except (KeyError, ValueError) as error:
            raise ClientError(response.text, response.status_code) from error



"""
Utilities for SAP Success Factors integrated channel.
"""


def populate_decrypted_fields_sap_success_factors(apps, schema_editor=None):  # pylint: disable=unused-argument
    """
    Populates the encryption fields in SAP Success Factors config with the data previously stored in database.
    """
    SAPSuccessFactorsEnterpriseCustomerConfiguration = apps.get_model(
        'sap_success_factors', 'SAPSuccessFactorsEnterpriseCustomerConfiguration'
    )

    for sap_success_factors_enterprise_configuration in SAPSuccessFactorsEnterpriseCustomerConfiguration.objects.all():
        sap_success_factors_enterprise_configuration.decrypted_key = getattr(
            sap_success_factors_enterprise_configuration, 'key', '')
        sap_success_factors_enterprise_configuration.decrypted_secret = getattr(
            sap_success_factors_enterprise_configuration, 'secret', '')
        sap_success_factors_enterprise_configuration.save()



"""
Views for the ``enterprise-customer-user`` API endpoint.
"""
from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import filters

from enterprise import models
from enterprise.api.filters import EnterpriseCustomerUserFilterBackend
from enterprise.api.pagination import PaginationWithFeatureFlags
from enterprise.api.v1 import serializers
from enterprise.api.v1.views.base_views import EnterpriseReadWriteModelViewSet


class EnterpriseCustomerUserViewSet(EnterpriseReadWriteModelViewSet):
    """
    API views for the ``enterprise-learner`` API endpoint.
    """

    queryset = models.EnterpriseCustomerUser.objects.all()
    filter_backends = (filters.OrderingFilter, DjangoFilterBackend, EnterpriseCustomerUserFilterBackend)
    pagination_class = PaginationWithFeatureFlags

    FIELDS = (
        'enterprise_customer', 'user_id', 'active',
    )
    filterset_fields = FIELDS
    ordering_fields = FIELDS

    def get_serializer_class(self):
        """
        Use a flat serializer for any requests that aren't read-only.
        """
        if self.request.method in ('GET',):
            return serializers.EnterpriseCustomerUserReadOnlySerializer

        return serializers.EnterpriseCustomerUserWriteSerializer



"""
Backfill missing audit record foreign keys for Cornerstone.
"""
import logging

from django.core.management.base import BaseCommand
from django.db.models import Q
from django.utils.translation import gettext as _

from integrated_channels.cornerstone.models import (
    CornerstoneEnterpriseCustomerConfiguration,
    CornerstoneLearnerDataTransmissionAudit,
)
from integrated_channels.integrated_channel.management.commands import IntegratedChannelCommandMixin
from integrated_channels.utils import batch_by_pk

LOGGER = logging.getLogger(__name__)


class Command(IntegratedChannelCommandMixin, BaseCommand):
    """
    Management command which backfills missing audit record foreign keys for cornerstone.
    """
    help = _('''
    Backfill missing audit record foreign keys for Cornerstone.
    ''')

    def find_csod_config_by_subdomain(self, customer_subdomain):
        """
        Given the subdomain from a CornerstoneLearnerDataTransmissionAudit record, search the
        CornerstoneEnterpriseCustomerConfiguration records for one with a matching base_url.
        Raise an exception when we dont find exactly one (missing config or duplicate configs are bad)
        """
        # real prod data often has a config like `https://edx.csod.com` alongside `https://edx-stg.csod.com`
        # if we just did a plain `icontains` using `edx` subdomain, we'd get the staging config too
        # expanding the subdomain into a proper url prefix lets us get a more exact match.
        # we require these urls be https
        subdomain_formatted_as_url_prefix = f'https://{customer_subdomain}.'
        configs = CornerstoneEnterpriseCustomerConfiguration.objects.filter(
            cornerstone_base_url__icontains=subdomain_formatted_as_url_prefix,
        )
        if len(configs) > 1:
            LOGGER.error(f'multiple ({len(configs)}) configs found for "{customer_subdomain}" when expecting just 1')
            return None
        else:
            return configs.first()

    def backfill_join_keys(self):
        """
        For each audit record kind, find all the records in batch, then lookup the appropriate
        enterprise_customer_uuid and/or plugin_config_id
        """
        try:
            only_missing_ld_fks = Q(plugin_configuration_id__isnull=True)
            for audit_record_batch in batch_by_pk(CornerstoneLearnerDataTransmissionAudit, extra_filter=only_missing_ld_fks):  # pylint: disable=line-too-long
                for audit_record in audit_record_batch:
                    config = self.find_csod_config_by_subdomain(audit_record.subdomain)
                    if config is None:
                        LOGGER.error(f'cannot find CSOD config for subdomain "{audit_record.subdomain}"')
                        continue
                    ent_cust_uuid = config.enterprise_customer.uuid
                    LOGGER.info(f'CornerstoneLearnerDataTransmissionAudit <{audit_record.pk}> '
                                f'enterprise_customer_uuid={ent_cust_uuid}, '
                                f'plugin_configuration_id={config.id}')
                    audit_record.enterprise_customer_uuid = ent_cust_uuid
                    audit_record.plugin_configuration_id = config.id
                    audit_record.save()
        except Exception as exc:
            LOGGER.exception('backfill_missing_csod_foreign_keys failed', exc_info=exc)
            raise exc

    def handle(self, *args, **options):
        """
        Backfill missing audit record foreign keys.
        """
        self.backfill_join_keys()



"""
    url mappings for integrated_channels/v1/moodle/
"""

from rest_framework import routers

from .views import MoodleConfigurationViewSet

app_name = 'moodle'
router = routers.DefaultRouter()
router.register(r'configuration', MoodleConfigurationViewSet, basename="configuration")
urlpatterns = router.urls



"""
Client for communicating with the E-Commerce API.
"""

import logging
from urllib.parse import urljoin

from requests.exceptions import ConnectionError, RequestException, Timeout  # pylint: disable=redefined-builtin

from django.conf import settings
from django.utils.translation import gettext as _

from enterprise.api_client.client import NoAuthAPIClient
from enterprise.utils import NotConnectedToOpenEdX, format_price

try:
    from openedx.core.djangoapps.site_configuration import helpers as configuration_helpers
except ImportError:
    configuration_helpers = None

try:
    from openedx.core.djangoapps.commerce.utils import get_ecommerce_api_client
except ImportError:
    get_ecommerce_api_client = None

LOGGER = logging.getLogger(__name__)


class EcommerceApiClient:
    """
    The API client to make calls to the E-Commerce API.
    """
    def __init__(self, user):
        """
        Create an E-Commerce API client, authenticated with the API token.

        This method retrieves an authenticated API client that can be used
        to access the ecommerce API. It raises an exception to be caught at
        a higher level if the package doesn't have OpenEdX resources available.
        """
        if get_ecommerce_api_client is None or configuration_helpers is None:
            raise NotConnectedToOpenEdX(
                _('To get a ecommerce_api_client, this package must be '
                  'installed in an Open edX environment.')
            )

        self.user = user
        self.client = get_ecommerce_api_client(user)
        self.API_BASE_URL = configuration_helpers.get_value(  # pylint: disable=invalid-name
            'ECOMMERCE_API_URL', settings.ECOMMERCE_API_URL
        )

    def get_course_final_price(self, mode, currency='$', enterprise_catalog_uuid=None):
        """
        Get course mode's SKU discounted price after applying any entitlement available for this user.

        Returns:
            str: Discounted price of the course mode.
        """
        api_url = urljoin(f"{self.API_BASE_URL}/", "baskets/calculate/")
        try:
            response = self.client.get(
                api_url,
                params={
                    "sku": [mode['sku']],
                    "username": self.user.username,
                    "catalog": enterprise_catalog_uuid,
                }
            )
            response.raise_for_status()
            price_details = response.json()
        except (RequestException, ConnectionError, Timeout) as exc:
            LOGGER.exception('Failed to get price details for sku %s due to: %s', mode['sku'], str(exc))
            price_details = {}
        price = price_details.get('total_incl_tax', mode['min_price'])
        if price != mode['min_price']:
            return format_price(price, currency)
        return mode['original_price']

    def create_manual_enrollment_orders(self, enrollments):
        """
        Calls ecommerce to create orders for the manual enrollments passed in.

        `enrollments` should be a list of enrollments with the following format::

            {
                "lms_user_id": <int>,
                "username": <str>,
                "email": <str>,
                "course_run_key": <str>
            }

        Since `student.CourseEnrollment` lives in LMS, we're just passing
        around dicts of the relevant information.
        """
        api_url = urljoin(f"{self.API_BASE_URL}/", "manual_course_enrollment_order/")
        try:
            response = self.client.post(api_url, json={"enrollments": enrollments})
            response.raise_for_status()
            order_response = response.json()
            order_creations = order_response["orders"]
            successful_creations = [order for order in order_creations if order["status"] == "success"]
            failed_creations = [order for order in order_creations if order["status"] == "failure"]
            if successful_creations:
                LOGGER.info(
                    "Successfully created orders for the following manual enrollments. %s",
                    successful_creations
                )
            if failed_creations:
                LOGGER.error(
                    "Failed to created orders for the following manual enrollments. %s",
                    failed_creations
                )
        except (RequestException, ConnectionError, Timeout) as exc:
            LOGGER.exception(
                "Failed to create order for manual enrollments for the following enrollments: %s. Reason: %s",
                enrollments,
                str(exc)
            )


class NoAuthEcommerceClient(NoAuthAPIClient):
    """
    The E-Commerce API client without authentication.
    """
    API_BASE_URL = settings.ECOMMERCE_PUBLIC_URL_ROOT
    APPEND_SLASH = False



# Generated by Django 3.2.19 on 2023-07-19 16:21

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('canvas', '0030_auto_20230112_2002'),
    ]

    operations = [
        migrations.AddField(
            model_name='canvasenterprisecustomerconfiguration',
            name='show_course_price',
            field=models.BooleanField(default=False, help_text='Displays course price'),
        ),
        migrations.AddField(
            model_name='historicalcanvasenterprisecustomerconfiguration',
            name='show_course_price',
            field=models.BooleanField(default=False, help_text='Displays course price'),
        ),
    ]



# Generated by Django 4.2.10 on 2024-06-13 10:47

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('enterprise', '0211_enterprisecustomer_show_videos_in_learner_portal_search_results_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='enterprisecustomer',
            name='career_engagement_network_message',
            field=models.TextField(blank=True, help_text='Message text shown on the learner portal dashboard for career engagement network.', null=True),
        ),
        migrations.AlterField(
            model_name='enterprisecustomer',
            name='learner_portal_sidebar_content',
            field=models.TextField(blank=True, help_text='Text shown on the learner portal dashboard for customer specific purposes. Open HTML field.', null=True),
        ),
        migrations.AlterField(
            model_name='historicalenterprisecustomer',
            name='career_engagement_network_message',
            field=models.TextField(blank=True, help_text='Message text shown on the learner portal dashboard for career engagement network.', null=True),
        ),
        migrations.AlterField(
            model_name='historicalenterprisecustomer',
            name='learner_portal_sidebar_content',
            field=models.TextField(blank=True, help_text='Text shown on the learner portal dashboard for customer specific purposes. Open HTML field.', null=True),
        ),
    ]



# Generated by Django 4.2.15 on 2024-10-07 14:33

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('consent', '0006_alter_historicaldatasharingconsent_options'),
    ]

    operations = [
        migrations.AddIndex(
            model_name='datasharingconsent',
            index=models.Index(fields=['username'], name='consent_dat_usernam_fae23a_idx'),
        ),
    ]



"""
Viewsets for integrated_channels/v1/logs
"""
import logging
from collections import OrderedDict
from datetime import datetime

from rest_framework import exceptions, permissions, viewsets
from rest_framework.pagination import PageNumberPagination
from rest_framework.response import Response
from rest_framework.status import (
    HTTP_200_OK,
    HTTP_201_CREATED,
    HTTP_202_ACCEPTED,
    HTTP_400_BAD_REQUEST,
    HTTP_404_NOT_FOUND,
    HTTP_409_CONFLICT,
    HTTP_422_UNPROCESSABLE_ENTITY,
    HTTP_500_INTERNAL_SERVER_ERROR,
)

from django.db.models import DateTimeField
from django.db.models.functions import Cast, Coalesce, Greatest

from integrated_channels.api.v1.mixins import PermissionRequiredForIntegratedChannelMixin
from integrated_channels.integrated_channel.models import (
    ContentMetadataItemTransmission,
    EnterpriseCustomerPluginConfiguration,
    LearnerDataTransmissionAudit,
)

from .serializers import ContentSyncStatusSerializer, LearnerSyncStatusSerializer

LOGGER = logging.getLogger(__name__)


class ReportingSyncStatusPagination(PageNumberPagination):
    page_size = 10
    page_size_query_param = 'page_size'

    def get_paginated_response(self, data):
        return Response(OrderedDict([
            ('pages_count', self.page.paginator.num_pages),
            ('count', self.page.paginator.count),
            ('next', self.get_next_link()),
            ('previous', self.get_previous_link()),
            ('results', data)
        ]))


class BaseReportingSyncStatusException(exceptions.APIException):
    status_code = 400


class InvalidSortByParamException(BaseReportingSyncStatusException):
    default_detail = "Invalid sort_by filter."
    default_code = "validation_error"


class InvalidChannelCodeException(BaseReportingSyncStatusException):
    default_detail = "Invalid channel code."
    default_code = "validation_error"


def validate_sort_by_query_params(model_table, request):
    if sort_by_param := request.GET.get('sort_by'):
        expected_sort_by_filters = ["sync_status", "sync_last_attempted_at"] + \
            [field.name for field in model_table._meta.fields]
        if sort_by_param.replace("-", "") not in expected_sort_by_filters:
            raise InvalidSortByParamException


def sort_by_date_aggregation(objects, date_fields, filter, order):
    # Arbitrarily old date to make sure null dates are considered lower in the `Greatest` annotation than any existing or to be
    # created datetime in the audits
    long_ago = datetime(year=1995, month=6, day=18)
    return objects.annotate(
        newest_date=Greatest(
            *[Coalesce(field, Cast(long_ago, DateTimeField()), output_field=DateTimeField()) for field in date_fields]
        )
    ).filter(**filter).order_by(f'{order}newest_date')


class ContentSyncStatusViewSet(PermissionRequiredForIntegratedChannelMixin, viewsets.ModelViewSet):
    """
    Sync-status APIs for `ContentMetadataItemTransmission` items.
    """
    serializer_class = ContentSyncStatusSerializer
    permission_classes = (permissions.IsAuthenticated,)
    permission_required = 'enterprise.can_access_admin_dashboard'
    pagination_class = ReportingSyncStatusPagination

    def validate(self):
        integrated_channel_code = self.kwargs.get('integrated_channel_code')
        channel_config_cls = EnterpriseCustomerPluginConfiguration.get_class_by_channel_code(integrated_channel_code)
        if channel_config_cls is None:
            raise InvalidChannelCodeException
        validate_sort_by_query_params(ContentMetadataItemTransmission, self.request)

    def get_queryset(self):
        self.validate()

        content_metadata_fields = ('enterprise_customer_uuid', 'integrated_channel_code', 'plugin_configuration_id')
        content_metadata_transmission_audit_filter = {field: self.kwargs.get(field) for field in content_metadata_fields}
        content_metadata_transmission_audit_filter['enterprise_customer_id'] = \
            content_metadata_transmission_audit_filter.pop('enterprise_customer_uuid')

        table_fields = [field.name for field in ContentMetadataItemTransmission._meta.fields]
        for query_filter in self.request.GET.keys():
            if query_filter in table_fields:
                content_metadata_transmission_audit_filter[f'{query_filter}__contains'] = self.request.GET.get(query_filter)

        # Sorting defaults to the status code field in DESC order to put failed transmission on top
        content_sort_by = ('-api_response_status_code',)
        if self.request.GET.get('sort_by'):
            # Translate the serialized `sync_status` to its DB field equivalent `api_response_status_code`
            content_sort_by = (self.request.GET.get('sort_by').replace('sync_status', 'api_response_status_code'),)

            # The serialized `sync_last_attempted_at` value is the representation of the `remote_<X>_at` values
            if "sync_last_attempted_at" in content_sort_by[0]:
                order = "-" if "-" in content_sort_by[0] else ""
                return sort_by_date_aggregation(
                    ContentMetadataItemTransmission.objects,
                    ['remote_updated_at', 'remote_deleted_at', 'remote_created_at'],
                    content_metadata_transmission_audit_filter,
                    order
                )

        return ContentMetadataItemTransmission.objects.filter(
            **content_metadata_transmission_audit_filter
        ).order_by(*content_sort_by)


class LearnerSyncStatusViewSet(PermissionRequiredForIntegratedChannelMixin, viewsets.ModelViewSet):
    """
    Sync-status APIs for `LearnerDataTransmissionAudit` implementation items.
    """
    permission_classes = (permissions.IsAuthenticated,)
    permission_required = 'enterprise.can_access_admin_dashboard'
    pagination_class = ReportingSyncStatusPagination

    def get_serializer_class(self):
        integrated_channel_code = self.kwargs.get('integrated_channel_code', None)
        serializer = LearnerSyncStatusSerializer.get_completion_class_by_channel_code(integrated_channel_code)
        if serializer is None:
            raise exceptions.ParseError("Invalid channel code.")
        return serializer

    def validate(self):
        integrated_channel_code = self.kwargs.get('integrated_channel_code', None)
        ThisLearnerClass = LearnerDataTransmissionAudit.get_completion_class_by_channel_code(integrated_channel_code)
        if ThisLearnerClass is None:
            raise InvalidChannelCodeException
        validate_sort_by_query_params(ThisLearnerClass, self.request)

    def get_queryset(self):
        self.validate()
        integrated_channel_code = self.kwargs.get('integrated_channel_code', None)
        ThisLearnerClass = LearnerDataTransmissionAudit.get_completion_class_by_channel_code(integrated_channel_code)

        learner_data_fields = ('enterprise_customer_uuid', 'plugin_configuration_id')
        learner_data_transmission_audit_filter = {field: self.kwargs.get(field) for field in learner_data_fields}

        table_fields = [field.name for field in ThisLearnerClass._meta.fields]
        for query_filter in self.request.GET.keys():
            if query_filter in table_fields:
                learner_data_transmission_audit_filter[f'{query_filter}__contains'] = self.request.GET.get(query_filter)

        # Defaults to DESC status code to put failed transmission on top
        content_sort_by = ('-status',)
        if self.request.GET.get('sort_by'):
            # Translate the serialized `sync_status` to its DB field equivalent `api_response_status_code`
            content_sort_by = (self.request.GET.get('sort_by').replace('sync_status', 'status'),)

            # The serialized `sync_last_attempted_at` value is the representation of both the `modified` and `created` values
            if "sync_last_attempted_at" in content_sort_by[0]:
                order = "-" if "-" in content_sort_by[0] else ""
                return sort_by_date_aggregation(
                    ThisLearnerClass.objects,
                    ['modified', 'created'],
                    learner_data_transmission_audit_filter,
                    order
                )

        return ThisLearnerClass.objects.filter(**learner_data_transmission_audit_filter).order_by(*content_sort_by)



# Generated by Django 3.2.9 on 2021-12-02 15:55

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('enterprise', '0149_invite_key_required_default_expiry_backfill'),
    ]

    operations = [
        migrations.AlterField(
            model_name='enterprisecustomerinvitekey',
            name='expiration_date',
            field=models.DateTimeField(help_text='The key will no longer be valid after this date.'),
        ),
        migrations.AlterField(
            model_name='historicalenterprisecustomerinvitekey',
            name='expiration_date',
            field=models.DateTimeField(help_text='The key will no longer be valid after this date.'),
        ),
    ]



# Generated by Django 1.11.20 on 2019-05-24 05:43


import simple_history.models

import django.db.models.deletion
import django.utils.timezone
from django.conf import settings
from django.db import migrations, models

import model_utils.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('enterprise', '0001_squashed_0092_auto_20200312_1650'),
    ]

    operations = [
        migrations.CreateModel(
            name='CornerstoneEnterpriseCustomerConfiguration',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('active', models.BooleanField(help_text='Is this configuration active?')),
                ('transmission_chunk_size', models.IntegerField(default=500, help_text='The maximum number of data items to transmit to the integrated channel with each request.')),
                ('cornerstone_base_url', models.CharField(blank=True, help_text='The base URL used for API requests to Cornerstone, i.e. https://portalName.csod.com', max_length=255, verbose_name='Cornerstone Base URL')),
                ('enterprise_customer', models.OneToOneField(help_text='Enterprise Customer associated with the configuration.', on_delete=django.db.models.deletion.CASCADE, to='enterprise.EnterpriseCustomer')),
            ],
        ),
        migrations.CreateModel(
            name='CornerstoneGlobalConfiguration',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('change_date', models.DateTimeField(auto_now_add=True, verbose_name='Change date')),
                ('enabled', models.BooleanField(default=False, verbose_name='Enabled')),
                ('completion_status_api_path', models.CharField(help_text='The API path for making completion POST requests to Cornerstone.', max_length=255, verbose_name='Completion Status API Path')),
                ('oauth_api_path', models.CharField(help_text='The API path for making OAuth-related POST requests to Cornerstone. This will be used to gain the OAuth access token which is required for other API calls.', max_length=255, verbose_name='OAuth API Path')),
                ('changed_by', models.ForeignKey(editable=False, null=True, on_delete=django.db.models.deletion.PROTECT, to=settings.AUTH_USER_MODEL, verbose_name='Changed by')),
            ],
        ),
        migrations.CreateModel(
            name='CornerstoneLearnerDataTransmissionAudit',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('user_guid', models.CharField(max_length=255)),
                ('enterprise_course_enrollment_id', models.PositiveIntegerField(blank=True, null=True)),
                ('course_id', models.CharField(help_text="The course run's key which is used to uniquely identify the course for Cornerstone.", max_length=255)),
                ('session_token', models.CharField(max_length=255)),
                ('callback_url', models.CharField(max_length=255)),
                ('subdomain', models.CharField(max_length=255)),
                ('course_completed', models.BooleanField(default=False, help_text="The learner's course completion status transmitted to Cornerstone.")),
                ('completed_timestamp', models.DateTimeField(blank=True, help_text='Date time when user completed course', null=True)),
                ('status', models.CharField(blank=True, max_length=100, null=True)),
                ('error_message', models.TextField(blank=True, null=True)),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='cornerstone_transmission_audit', to=settings.AUTH_USER_MODEL)),
            ],
        ),
        migrations.CreateModel(
            name='HistoricalCornerstoneEnterpriseCustomerConfiguration',
            fields=[
                ('id', models.IntegerField(auto_created=True, blank=True, db_index=True, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('active', models.BooleanField(help_text='Is this configuration active?')),
                ('transmission_chunk_size', models.IntegerField(default=500, help_text='The maximum number of data items to transmit to the integrated channel with each request.')),
                ('cornerstone_base_url', models.CharField(blank=True, help_text='The base URL used for API requests to Cornerstone, i.e. https://portalName.csod.com', max_length=255, verbose_name='Cornerstone Base URL')),
                ('history_id', models.AutoField(primary_key=True, serialize=False)),
                ('history_date', models.DateTimeField()),
                ('history_change_reason', models.CharField(max_length=100, null=True)),
                ('history_type', models.CharField(choices=[('+', 'Created'), ('~', 'Changed'), ('-', 'Deleted')], max_length=1)),
                ('enterprise_customer', models.ForeignKey(blank=True, db_constraint=False, help_text='Enterprise Customer associated with the configuration.', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='enterprise.EnterpriseCustomer')),
                ('history_user', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'ordering': ('-history_date', '-history_id'),
                'get_latest_by': 'history_date',
                'verbose_name': 'historical cornerstone enterprise customer configuration',
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.AlterUniqueTogether(
            name='cornerstonelearnerdatatransmissionaudit',
            unique_together={('user', 'course_id')},
        ),
    ]
